# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Benchmark module
"""

from .base_benchmark import BaseTrainingProcedure, MislabelingDetectorEvaluator
from .influence_factory import (
    InfluenceCalculatorFactory,
    FirstOrderFactory,
    RPSLJEFactory,
    TracInFactory,
    WeightsBoundaryCalculatorFactory,
    SampleBoundaryCalculatorFactory,
    ArnoldiCalculatorFactory
)
from .cifar10_benchmark import Cifar10TrainingProcedure, Cifar10MislabelingDetectorEvaluator



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module containing the base class for representer point theorem-based influence calculators
"""
from abc import abstractmethod

import tensorflow as tf
from tensorflow.keras import Model
from tensorflow.keras.losses import Loss, Reduction

from ..common import BaseInfluenceCalculator
from ..types import Tuple, Callable, Union

from ..utils import assert_batched_dataset, split_model


class BaseRepresenterPoint(BaseInfluenceCalculator):
    """
    Base interface for representer point theorem-based influence calculators.

    Disclaimer: This method only works on classification problems!

    Parameters
    ----------
    model
        A TF2 model that has already been trained
    train_set
        A batched TF dataset with the points with which the model was trained
    loss_function
        The loss function with which the model was trained. This loss function MUST NOT be reduced.
    """
    def __init__(
            self,
            model: Model,
            train_set: tf.data.Dataset,
            loss_function: Union[Callable[[tf.Tensor, tf.Tensor], tf.Tensor], Loss],
            target_layer: Union[str, int] = -1
    ):
        # Make sure that the dataset is batched and that the loss function is not reduced
        assert_batched_dataset(train_set)
        self.train_set = train_set
        if hasattr(loss_function, 'reduction'):
            assert loss_function.reduction == Reduction.NONE

        # Make sure that the model's last layer is a Dense layer with no bias
        if not isinstance(model.layers[-1], tf.keras.layers.Dense):
            raise ValueError('The last layer of the model must be a Dense layer with no bias.')
        if model.layers[-1].use_bias:
            raise ValueError('The last layer of the model must be a Dense layer with no bias.')
        self.loss_function = loss_function

        # Cut the model in two (feature extractor and head)
        self.model = model
        self.target_layer = target_layer
        self.feature_extractor, self.original_head = split_model(model, target_layer)

    @abstractmethod
    def _compute_alpha(self, z_batch: tf.Tensor, y_batch: tf.Tensor) -> tf.Tensor:
        """
        Compute the alpha vector for a given input-output pair (z, y)

        Parameters
        ----------
        z_batch
            A tensor containing the latent representation of an input point.
        y_batch
            The labels corresponding to the representations z

        Returns
        -------
        alpha
            A tensor with the alpha coefficients of the kernel given by the representer point theorem
        """
        raise NotImplementedError()

    def _preprocess_samples(self, samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Preprocess a single batch of samples.

        Parameters
        ----------
        samples
            A single batch of tensors containing the samples.

        Returns
        -------
        evaluate_vect
            The preprocessed sample
        """
        x_batch = self.feature_extractor(samples[:-1])
        y_t = samples[-1]

        return x_batch, y_t

    def _compute_influence_vector(self, train_samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Compute an equivalent of the influence vector for a sample of training points.

        Disclaimer: this vector is not an estimation of the difference between the actual
        model and the perturbed model without the samples (like it is the case with what is
        calculated using deel.influenciae.influence).

        Parameters
        ----------
        train_samples
            A tensor with a group of training samples of which we wish to compute the influence.

        Returns
        -------
        influence_vectors
            A tensor with a concatenation of the alpha weights and the feature maps for each sample.
            This allows for optimizations to be put in place but is not really an influence vector
            of any kind.
        """
        x_batch = self.feature_extractor(train_samples[:-1])
        alpha = self._compute_alpha(x_batch, train_samples[-1])

        return alpha, x_batch

    def _estimate_individual_influence_values_from_batch(
            self,
            train_samples: Tuple[tf.Tensor, ...],
            samples_to_evaluate: Tuple[tf.Tensor, ...]
    ) -> tf.Tensor:
        """
        Estimate the (individual) influence scores of a single batch of samples with respect to
        a batch of samples belonging to the model's training dataset.

        Parameters
        ----------
        train_samples
            A single batch of training samples (and their target values).
        samples_to_evaluate
            A single batch of samples of which we wish to compute the influence of removing the training
            samples.

        Returns
        -------
        A tensor containing the individual influence scores.
        """
        return self._estimate_influence_value_from_influence_vector(
            self._preprocess_samples(samples_to_evaluate),
            self._compute_influence_vector(train_samples)
        )

    def _estimate_influence_value_from_influence_vector(
            self,
            preproc_test_sample: tf.Tensor,
            influence_vector: tf.Tensor
    ) -> tf.Tensor:
        """
        Compute the influence score for a (batch of) preprocessed test sample(s) and a training "influence vector".

        Parameters
        ----------
        preproc_test_sample
            A tensor with a pre-processed sample to evaluate.
        influence_vector
            A tensor with the training influence vector.

        Returns
        -------
        influence_values
            A tensor with influence values for the (batch of) test samples.
        """
        # Extract the different information inside the tuples
        feature_maps_test, _ = preproc_test_sample
        alpha, feature_maps_train = influence_vector

        if len(alpha.shape) == 1 or (len(alpha.shape) == 2 and alpha.shape[1] == 1):
            influence_values = alpha * tf.matmul(feature_maps_train, feature_maps_test, transpose_b=True)
        else:
            influence_values = tf.gather(
                alpha, tf.argmax(self.original_head(feature_maps_test), axis=1), axis=1, batch_dims=1
            ) * tf.matmul(feature_maps_train, feature_maps_test, transpose_b=True)
        influence_values = tf.transpose(influence_values)

        return influence_values

    def _compute_influence_value_from_batch(self, train_samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Compute the influence score for a batch of training samples (i.e. self-influence).

        Parameters
        ----------
        train_samples
            A tensor containing a batch of training samples.

        Returns
        -------
        influence_values
            A tensor with the self-influence of the training samples.
        """
        x_batch = self.feature_extractor(train_samples[:-1])
        alpha = self._compute_alpha(x_batch, train_samples[-1])

        # If the problem is binary classification, take all the alpha values
        # If multiclass, take only those that correspond to the prediction
        out_shape = self.model.output_shape
        if len(out_shape) == 1:
            influence_values = alpha
        elif len(out_shape) == 2 and out_shape[1] == 1:
            influence_values = alpha
        else:
            if len(out_shape) > 2:
                indices = tf.argmax(tf.squeeze(self.original_head(x_batch), axis=-1), axis=1)
            else:
                indices = tf.argmax(self.original_head(x_batch), axis=1)
            influence_values = tf.gather(alpha, indices, axis=1, batch_dims=1)

        return tf.abs(influence_values)



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module implementing a technique based on the representer point theorem for kernels,
but using a local jacobian expansion, as per
https://proceedings.neurips.cc/paper/2021/file/c460dc0f18fc309ac07306a4a55d2fd6-Paper.pdf
"""
import tensorflow as tf

from .base_representer_point import BaseRepresenterPoint
from ..common import InfluenceModel, InverseHessianVectorProductFactory
from ..types import Union, Optional


class RepresenterPointLJE(BaseRepresenterPoint):
    """
    Representer Point Selection via Local Jacobian Expansion for Post-hoc Classifier Explanation of Deep Neural
    Networks and Ensemble Models
    https://proceedings.neurips.cc/paper/2021/file/c460dc0f18fc309ac07306a4a55d2fd6-Paper.pdf

    Disclaimer: This technique requires the last layer of the model to be a Dense layer with no bias.

    Parameters
    ----------
    influence_model
        The TF2.X model implementing the InfluenceModel interface.
    ihvp_calculator_factory
        An InverseHessianVectorProductFactory for creating new instances of the InverseHessianVectorProduct
        class.
    n_samples_for_hessian
        An integer for the amount of samples from the training dataset that will be used for the computation of the
        hessian matrix.
        If None, the whole dataset will be used.
    target_layer
        Either a string or an integer identifying the layer on which to compute the influence-related quantities.
    shuffle_buffer_size
        An integer with the buffer size for the training set's shuffle operation.
    epsilon
        An epsilon value to prevent division by zero.
    """
    def __init__(
            self,
            influence_model: InfluenceModel,
            dataset: tf.data.Dataset,
            ihvp_calculator_factory: InverseHessianVectorProductFactory,
            n_samples_for_hessian: Optional[int] = None,
            target_layer: Union[int, str] = -1,
            shuffle_buffer_size: int = 10000,
            epsilon: float = 1e-5
    ):
        super().__init__(influence_model.model, dataset, influence_model.loss_function)
        self.epsilon = tf.constant(epsilon, dtype=tf.float32)

        # In the paper, the authors explain that in practice, they use a single step of SGD to compute the
        # perturbed model's weights. We will do the same here.
        optimizer = tf.keras.optimizers.SGD(learning_rate=1e-4)
        target_layer_shape = influence_model.model.layers[target_layer].input.type_spec.shape
        perturbed_head = tf.keras.models.clone_model(self.original_head)
        perturbed_head.set_weights(self.original_head.get_weights())
        perturbed_head.build(target_layer_shape)
        perturbed_head.compile(optimizer=optimizer, loss=influence_model.loss_function)

        # Get a dataset to compute the SGD step
        if n_samples_for_hessian is None:
            dataset_to_estimate_hessian = dataset
        else:
            n_batches_for_hessian = max(n_samples_for_hessian // dataset._batch_size, 1)
            dataset_to_estimate_hessian = dataset.shuffle(shuffle_buffer_size).take(n_batches_for_hessian)
        f_array, y_array = None, None
        for x, y in dataset_to_estimate_hessian:
            f = self.feature_extractor(x)
            f_array = f if f_array is None else tf.concat([f_array, f], axis=0)
            y_array = y if y_array is None else tf.concat([y_array, y], axis=0)
        dataset_to_estimate_hessian = tf.data.Dataset.from_tensor_slices((f_array, y_array)).batch(dataset._batch_size)

        # Accumulate the gradients for the whole dataset and then update
        trainable_vars = perturbed_head.trainable_variables
        accum_vars = [tf.Variable(tf.zeros_like(t_var.read_value()), trainable=False)
                      for t_var in trainable_vars]
        for x, y in dataset_to_estimate_hessian:
            with tf.GradientTape() as tape:
                y_pred = perturbed_head(x)
                loss = -perturbed_head.loss(y, y_pred)
            gradients = tape.gradient(loss, trainable_vars)
            _ = [accum_vars[i].assign_add(grad) for i, grad in enumerate(gradients)]
        optimizer.apply_gradients(zip(accum_vars, trainable_vars))

        # Keep the perturbed head
        self.perturbed_head = perturbed_head

        # Create the new model with the perturbed weights to compute the hessian matrix
        model = InfluenceModel(
            self.perturbed_head,
            1,  # layer 0 is InputLayer
            loss_function=influence_model.loss_function
        )
        self.ihvp_calculator = ihvp_calculator_factory.build(model, dataset_to_estimate_hessian)

    def _compute_alpha(self, z_batch: tf.Tensor, y_batch: tf.Tensor) -> tf.Tensor:
        """
        Computes the alpha vector for the Local Jacobian Expansion approximation.

        Parameters
        ----------
        z_batch
            A tensor with the perturbed model's predictions.
        y_batch
            A tensor with the ground truth labels.

        Returns
        -------
        A tensor with the alpha vector for the Local Jacobian Expansion approximation.
        """
        # First, we compute the second term, which contains the Hessian vector product
        weights = self.perturbed_head.trainable_weights
        with tf.GradientTape(persistent=False, watch_accessed_variables=False) as tape:
            tape.watch(weights)
            logits = self.perturbed_head(z_batch)
            loss = self.perturbed_head.compiled_loss(y_batch, logits)
        grads = tape.jacobian(loss, weights)[0]
        grads = tf.multiply(
            grads,
            tf.repeat(
                tf.expand_dims(
                    tf.divide(tf.ones_like(z_batch),
                              tf.cast(tf.shape(z_batch)[0], z_batch.dtype) * z_batch +
                              tf.cast(self.epsilon, z_batch.dtype)),
                    axis=-1),
                grads.shape[-1], axis=-1
            )
        )
        second_term = tf.map_fn(
            lambda v: self.ihvp_calculator._compute_ihvp_single_batch(  # pylint: disable=protected-access
                tf.expand_dims(v, axis=0),
                use_gradient=False
            ),
            grads
        )
        second_term = tf.reduce_sum(tf.reshape(second_term, tf.shape(grads)), axis=1)

        # Second, we compute the first term, which contains the weights
        first_term = tf.concat(list(weights), axis=0)
        first_term = tf.multiply(
            first_term,
            tf.repeat(
                tf.expand_dims(
                    tf.divide(tf.ones_like(z_batch),
                              tf.cast(tf.shape(z_batch)[0], z_batch.dtype) * z_batch +
                              tf.cast(self.epsilon, z_batch.dtype)),
                    axis=-1),
                first_term.shape[-1], axis=-1
            )
        )
        first_term = tf.reduce_sum(first_term, axis=1)

        return first_term - second_term  # alpha is first term minus second term



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module defining the interface and classes that implement factories for objects of
class InverseHessianVectorProduct. This will prove itself useful for creating
the objects necessary for computing the different (I)HVPs in second order influence
functions.
"""
from abc import abstractmethod

import tensorflow as tf

from .model_wrappers import InfluenceModel
from .inverse_hessian_vector_product import (
    InverseHessianVectorProduct,
    ExactIHVP,
    ConjugateGradientDescentIHVP,
    LissaIHVP
)

from ..types import Union, Optional


class InverseHessianVectorProductFactory:
    """
    The base interface for InverseHessianVectorProduct factories.
    """
    @abstractmethod
    def build(self, model_influence: InfluenceModel, dataset: tf.data.Dataset) -> InverseHessianVectorProduct:
        """
        Creates an instance of an InverseHessianVectorProduct class with the provided
        parameters.

        Parameters
        ----------
        model_influence
            A TF model implementing the InfluenceModel interface.
        dataset
            A TF dataset object containing the model's (full or partial) training dataset.

        Returns
        -------
        ihvp
            An instance of an InverseHessianVectorProduct class.
        """
        raise NotImplementedError()


class ExactIHVPFactory(InverseHessianVectorProductFactory):
    """
    A factory for instantiating ExactIHVP objects.
    """
    def build(self, model_influence: InfluenceModel, dataset: tf.data.Dataset) -> InverseHessianVectorProduct:
        """
        Creates an instance of the ExactIHVP class for a given model
        implementing the InfluenceModel interface and its (full or partial) training dataset.

        Parameters
        ----------
        model_influence
            A TF model implementing the InfluenceModel interface
        dataset
            A TF dataset containing the model's (full or partial) training dataset.

        Returns
        -------
        exact_ihvp
            An instance of the ExactIHVP class.
        """
        return ExactIHVP(model_influence, dataset)


class CGDIHVPFactory(InverseHessianVectorProductFactory):
    """
    A factory for instantiating ConjugateGradientDescentIHVP objects.

    Attributes
    ----------
    feature_extractor
        Either a TF feature-extractor model or the index of the layer of a whole model
        which will be cut into two for computing the influence vectors and scores.
    n_cgd_iters
        An integer specifying the amount of iterations of the optimizer to run before
        (prematurely) considering the optimization completed.
    extractor_layer
        The cutoff layer for the feature extractor, if specified in TF model format.
    """
    def __init__(
        self,
        feature_extractor: Union[int, tf.keras.Model] = -1,
        n_cgd_iters: int = 100,
        extractor_layer: Optional[Union[str, int]] = None
    ):
        self.n_cgd_iters = n_cgd_iters
        if isinstance(feature_extractor, int):
            self.extractor_layer = feature_extractor
            self.feature_extractor = None
        else:
            assert extractor_layer is not None, "If you provide a model as a feature extractor, you should also" \
                                                "provide the id of the last extracted layer"
            self.extractor_layer = extractor_layer
            self.feature_extractor = feature_extractor

    def build(self, model_influence: InfluenceModel, dataset: tf.data.Dataset) -> InverseHessianVectorProduct:
        """
        Creates an instance of the ConjugateGradientDescentIHVP class for the provided model and its
        corresponding (full or partial) training dataset.

        Parameters
        ----------
        model_influence
            A TF model implementing the InfluenceModel interface.
        dataset
            A TF dataset containing the model's (full or partial) training dataset.

        Returns
        -------
        cgd_ihvp
            An instance of the ConjugateGradientDescentIHVP class
        """
        return ConjugateGradientDescentIHVP(
            model_influence,
            self.extractor_layer,
            dataset,
            self.n_cgd_iters,
            self.feature_extractor,
        )


class LissaIHVPFactory(InverseHessianVectorProductFactory):
    """
    A factory for instantiating LissaIHVP objects.

    Attributes
    ----------
    feature_extractor
        Either a TF feature-extractor model or the index of the layer of a whole model
        which will be cut into two for computing the influence vectors and scores.
    n_cgd_iters
        An integer specifying the amount of iterations of the optimizer to run before
        (prematurely) considering the optimization completed.
    extractor_layer
        The cutoff layer for the feature extractor, if specified in TF model format.
    damping
        A damping parameter to regularize a nearly singular operator.
    scale
        A rescaling factor to verify the hypothesis of norm(operator / scale) < 1.
    """
    def __init__(
        self,
        feature_extractor: Union[int, tf.keras.Model] = -1,
        n_cgd_iters: int = 100,
        extractor_layer: Optional[Union[str, int]] = None,
        damping: float = 1e-4,
        scale: float = 10.
    ):
        self.n_cgd_iters = n_cgd_iters
        self.damping = damping
        self.scale = scale
        if isinstance(feature_extractor, int):
            self.extractor_layer = feature_extractor
            self.feature_extractor = None
        else:
            assert extractor_layer is not None, "If you provide a model as a feature extractor, you should also" \
                                                "provide the id of the last extracted layer"
            self.extractor_layer = extractor_layer
            self.feature_extractor = feature_extractor

    def build(self, model_influence: InfluenceModel, dataset: tf.data.Dataset) -> InverseHessianVectorProduct:
        """
        Creates an instance of the ConjugateGradientDescentIHVP class for the provided model and its
        corresponding (full or partial) training dataset.

        Parameters
        ----------
        model_influence
            A TF model implementing the InfluenceModel interface.
        dataset
            A TF dataset containing the model's (full or partial) training dataset.

        Returns
        -------
        cgd_ihvp
            An instance of the ConjugateGradientDescentIHVP class
        """
        return LissaIHVP(
            model_influence,
            self.extractor_layer,
            dataset,
            self.n_cgd_iters,
            self.feature_extractor,
            self.damping,
            self.scale
        )



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Visualizations module
"""

from .benchmark import BenchmarkDisplay
from .image import plot_most_influential_images, plot_datacentric_explanations



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module containing the implementation of a SortedDictionary, useful for computing the
top-k most influential samples in a large dataset.
"""
from enum import Enum

import tensorflow as tf
import numpy as np


class ORDER(Enum):
    """
    Enumeration for the two types of ordering for the sorting function.
    ASCENDING puts the elements with the smallest value first.
    DESCENDING puts the elements with the largest value first.
    """
    ASCENDING = 1
    DESCENDING = 2


class BatchSort:
    """
    An implementation of a SortedDictionary that accepts batches of elements and their corresponding scores
    and keeps only the k most/least important ones.

    Attributes
    ----------
    batch_shape
        A tuple with the shape of the dictionary's inputs.
    k_shape
        A tuple with the shape of the elements to keep.
    dtype
        The data-type of the input's values. By default, float32 will be used.
    order
        Either ASCENDING or DESCENDING depending on whether the most or the least important elements are to
        be kept.
    """
    def __init__(self, batch_shape, k_shape, dtype=tf.float32, order: ORDER = ORDER.DESCENDING):
        self.k = k_shape[1]
        shape = tf.concat((k_shape, batch_shape), axis=0)
        self.best_batch = tf.Variable(tf.zeros(shape, dtype=dtype), trainable=False)
        self.order = order
        if self.order == ORDER.DESCENDING:
            self.best_values = tf.Variable(tf.ones(k_shape, dtype=dtype) * (-np.inf), trainable=False)
        else:
            self.best_values = tf.Variable(tf.ones(k_shape, dtype=dtype) * np.inf, trainable=False)

    def add_all(self, batch_key: tf.Tensor, batch_values: tf.Tensor) -> None:
        """
        Add a new batch of data (element and values) and update the sorted dictionary retaining only the
        top/bottom-k elements.

        Parameters
        ----------
        batch_key
            A batch of new elements in the form of a tensor.
        batch_values
            A batch of their corresponding values in the form of a tensor.
        """
        current_score = tf.concat([self.best_values, batch_values], axis=1)
        current_batch = tf.concat([self.best_batch, batch_key], axis=1)

        if self.order == ORDER.DESCENDING:
            indexes = tf.argsort(current_score, axis=1, direction='DESCENDING')
        else:
            indexes = tf.argsort(current_score, axis=1, direction='ASCENDING')

        indexes = indexes[:, :self.k]

        current_best_score = tf.gather(current_score, indexes, batch_dims=1)
        current_best_samples = tf.gather(current_batch, indexes, batch_dims=1)

        self.best_values.assign(current_best_score)
        self.best_batch.assign(current_best_samples)

    def get(self):
        """
        A getter method for the top-k elements and their corresponding scores.

        Returns
        -------
        top_k_tuple
            A tuple with the top_k elements and their scores.
        """
        return self.best_batch, self.best_values

    def reset(self):
        """
        Resets the values of the whole object and prepares it to start sorting from scratch.
        """
        self.best_batch.assign(tf.zeros_like(self.best_batch))

        if self.order == ORDER.DESCENDING:
            self.best_values.assign((-np.inf) * tf.ones_like(self.best_values))
        else:
            self.best_values.assign(np.inf * tf.ones_like(self.best_values))



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module implementing a second order approximation for groups of data-points, where
pairwise interactions of how holding out one sample affects another in the group
that's being held out are now taken into account. This method was originally
presented in https://arxiv.org/abs/1911.00418 and it is supposed to greatly improve
the accuracy of the influence estimations for big groups of data.

Disclaimer: this method can be very computationally expensive, especially when calculating
the influence for a large number of weights.
"""
import tensorflow as tf

from .base_group_influence import BaseGroupInfluenceCalculator
from ..common import ExactIHVP, ConjugateGradientDescentIHVP, LissaIHVP
from ..common import InfluenceModel
from ..common import InverseHessianVectorProduct, IHVPCalculator

from ..utils import assert_batched_dataset, dataset_size
from ..types import Optional, Union


class SecondOrderInfluenceCalculator(BaseGroupInfluenceCalculator):
    """
    A class implementing the necessary methods to compute the different influence quantities
    (only for groups) using a second-order approximation, thus allowing us to take into
    account the pairwise interactions between points inside the group. For small groups of
    points, consider using the first order alternative if the computational cost is
    too high.

    Notes
    -----
    The methods currently implemented are available to evaluate groups of points:
    - Influence function vectors: the weights difference when removing groups of points.
    - Influence values/Cook's distance: a measure of reliance of the model on the group of points.

    This implementation is based on the following paper:
    [https://arxiv.org/abs/1911.00418](https://arxiv.org/abs/1911.00418)

    Parameters
    ----------
    model
        The TF2.X model implementing the InfluenceModel interface.
    dataset
        A batched TF dataset containing the training dataset over which we will estimate the
        inverse-hessian-vector product.
    ihvp_calculator
        Either a string containing the IHVP method ('exact' or 'cgd'), an IHVPCalculator
        object or an InverseHessianVectorProduct object.
    n_samples_for_hessian
        An integer indicating the amount of samples to take from the provided train dataset.
    shuffle_buffer_size
        An integer indicating the buffer size of the train dataset's shuffle operation -- when
        choosing the amount of samples for the hessian.
    """
    def __init__(
            self,
            model: InfluenceModel,
            dataset: tf.data.Dataset,
            ihvp_calculator: Union[str, InverseHessianVectorProduct, IHVPCalculator] = 'exact',
            n_samples_for_hessian: Optional[int] = None,
            shuffle_buffer_size: Optional[int] = 10000
    ):

        super().__init__(
            model,
            dataset,
            ihvp_calculator,
            n_samples_for_hessian,
            shuffle_buffer_size
        )

        self.train_size = dataset_size(dataset)

    def compute_influence_vector_group(
            self,
            group: tf.data.Dataset
    ) -> tf.Tensor:
        """
        Computes the influence function vector -- an estimation of the weights difference when
        removing the points -- of the whole group of points.

        Parameters
        ----------
        group
            A batched TF dataset containing the group of points of which we wish to compute the
            influence of removal.

        Returns
        -------
        influence_group
            A tensor containing one vector for the whole group.
        """
        assert_batched_dataset(group)
        fraction = tf.cast(dataset_size(group), dtype=tf.float32) / tf.cast(self.train_size, dtype=tf.float32)

        coeff_additive_term = (1. - 2 * fraction) / \
                              (tf.square(1. - fraction) * tf.cast(self.train_size, dtype=tf.float32))
        coeff_pairwise_term = 1. / tf.square((1. - fraction) * tf.cast(self.train_size, dtype=tf.float32))

        additive = coeff_additive_term * self._compute_additive_term(group)
        pairwise = coeff_pairwise_term * self._compute_pairwise_interactions(group)

        influence_group = additive + pairwise
        influence_group = tf.transpose(influence_group)  # to get the right shape for the output

        return influence_group

    def _compute_additive_term(self, dataset: tf.data.Dataset):
        """
        Computes the additive term as per Basu et al.'s article. It accounts for the influence of each of the
        points that we wish to remove, without taking into account the interactions between each other

        Parameters
        ----------
        dataset
            A batched TF dataset containing the points we wish to remove

        Returns
        -------
        interactions
            A tensor containing the addition of the influence of the points in the group
        """
        ihvp_ds = self.ihvp_calculator.compute_ihvp(dataset)
        reduced_ihvp = ihvp_ds.map(lambda x: tf.reduce_sum(x, axis=1, keepdims=True))
        return reduced_ihvp.reduce(tf.constant(0, dtype=ihvp_ds.element_spec.dtype), lambda x, y: x + y)

    def _compute_pairwise_interactions(self, dataset: tf.data.Dataset):
        """
        Computes the term corresponding to the pairwise interactions as per Basu et al.'s article. It will
        contain all the interactions between each of the points with each of the other points.

        Disclaimer: this term can be quite computationally intensive to calculate, as there are 3 nested hessian-vector
        products. Thus, it can also be a source of numerical instability when using an approximate version of an
        IHVP calculator.

        Parameters
        ----------
        dataset
            A batched TF dataset containing the points we wish to remove

        Returns
        -------
        interactions
            A tensor containing the sum of all the interactions of each point we are removing with each other point
            of the group
        """
        if isinstance(self.ihvp_calculator, ExactIHVP):
            local_ihvp = ExactIHVP(self.model, dataset)
        elif isinstance(self.ihvp_calculator, ConjugateGradientDescentIHVP):
            local_ihvp = ConjugateGradientDescentIHVP(self.model, self.ihvp_calculator.extractor_layer,
                                                      dataset, self.ihvp_calculator.n_opt_iters,
                                                      self.ihvp_calculator.feature_extractor)
        else:
            local_ihvp = LissaIHVP(self.model, self.ihvp_calculator.extractor_layer,
                                   dataset, self.ihvp_calculator.n_opt_iters,
                                   self.ihvp_calculator.feature_extractor)

        ihvp_ds = self.ihvp_calculator.compute_ihvp(dataset)
        ihvp_ds = ihvp_ds.map(lambda x: tf.reduce_sum(x, axis=1))

        reduced_ihvp = ihvp_ds.reduce(tf.constant(0, dtype=ihvp_ds.element_spec.dtype), lambda x, y: x + y)
        reduced_ihvp_ds = tf.data.Dataset.from_tensors(reduced_ihvp).batch(dataset._batch_size)  # pylint: disable=W0212

        local_hvp = local_ihvp.compute_hvp(reduced_ihvp_ds, use_gradient=False).batch(
            dataset._batch_size)  # pylint: disable=W0212

        interactions = self.ihvp_calculator.compute_ihvp(
            local_hvp, use_gradient=False
        )

        ds_size = tf.cast(dataset_size(dataset), dtype=interactions.element_spec.dtype)
        interactions = interactions.map(lambda x: x * ds_size)

        return interactions.get_single_element()

    def estimate_influence_values_group(
            self,
            group_train: tf.data.Dataset,
            group_to_evaluate: Optional[tf.data.Dataset] = None
    ) -> tf.Tensor:
        """
        Computes Cook's distance of the whole group of points provided, giving measure of the
        influence that the group carries on the model's weights.

        The dataset_train contains the points we will be removing and dataset_to_evaluate,
        those with respect to whom we will be measuring the influence. As we will be performing
        the same operation in batches, we consider that each point from one dataset corresponds
        to one from the other. As such, both datasets must contain the same amount of points.
        In case the group_to_evaluate is not given, use by default the
        group_to_train: compute the self influence of the group.


        Parameters
        ----------
        group_train
            A batched TF dataset containing the group of points we wish to remove.
        group_to_evaluate
            A batched TF dataset containing the group of points with respect to whom we wish to
            measure the influence of removing the training points.

        Returns
        -------
        influence_values_group
            A tensor containing one influence value for the whole group.
        """
        if group_to_evaluate is None:
            group_to_evaluate = group_train
        ds_size = self.assert_compatible_datasets(group_train, group_to_evaluate)
        influence = tf.transpose(self.compute_influence_vector_group(group_train))
        reduced_grads = tf.reduce_sum(tf.reshape(self.model.batch_jacobian(group_to_evaluate),
                                                 (ds_size, -1)), axis=0, keepdims=True)

        return tf.matmul(reduced_grads, influence)



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Common classes and methods
"""

from .model_wrappers import BaseInfluenceModel, InfluenceModel
from .base_influence import SelfInfluenceCalculator, BaseInfluenceCalculator, CACHE
from .inverse_hessian_vector_product import InverseHessianVectorProduct, ExactIHVP, ConjugateGradientDescentIHVP, \
     IHVPCalculator, LissaIHVP, ForwardOverBackwardHVP
from .ihvp_factory import InverseHessianVectorProductFactory, ExactIHVPFactory, CGDIHVPFactory, LissaIHVPFactory



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module containing methods based on decision boundaries and adversarial attacks for
estimating the influence of data-points.
"""
from .weights_boundary import WeightsBoundaryCalculator
from .sample_boundary import SampleBoundaryCalculator



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Representer Point theorem module
"""

from .base_representer_point import BaseRepresenterPoint
from .rps_l2 import RepresenterPointL2
from .rps_lje import RepresenterPointLJE



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module implementing a data-point-influence estimation method based on how intensely
each sample must be adversarially attacked to make it change its prediction. Intuitively,
we would expect outliers and atypical examples to need a lesser attack for the model to
misclassify it.

Unlike other influence calculators, this one cannot be used to estimate
the influence of a point on another.

The adversarial attacks are performed via deep fool.
"""
import tensorflow as tf
from tensorflow.keras import Model  # pylint:  disable=E0611

from ..common import SelfInfluenceCalculator
from ..types import Tuple


class SampleBoundaryCalculator(SelfInfluenceCalculator):
    """
    A class implementing an influence score based on the distance of a sample to the
    boundary of the classifier.
    The distance to the boundary is estimated using the deep fool method.
    [https://arxiv.org/abs/1511.04599]

    Notes
    -----
    This method has better mislabeled-sample-detection performance when the model overfits.

    Parameters
    ----------
    model
        A TF2 model that has already been trained
    step_nbr
        Number of the iterations to find the closest adversarial problem
    eps
        Difference between two logits to assume that they have the same values
    """

    def __init__(self, model: Model, step_nbr: int = 100, eps: float = 1E-6):
        self.weights_init = [tf.identity(w) for w in model.trainable_variables]
        self.model = model

        self.step_nbr = step_nbr
        self.eps = eps

    @staticmethod
    def __delta_to_index(indexes_1: tf.Tensor, indexes_2: tf.Tensor, x: tf.Tensor):
        """
        Compute the difference between the logit of a given class and the other logits

        Parameters
        ----------
        indexes_1
            The logits of other classes
        indexes_2
            The logits of the predicted class
        x
            The logits

        Returns
        -------
        delta_x
            The difference between the logits
        """
        x1 = tf.gather(x, indexes_1, batch_dims=1)
        x2 = tf.gather(x, tf.expand_dims(indexes_2, axis=1), batch_dims=1)

        delta_x = x1 - tf.repeat(x2, tf.shape(x1)[1], axis=1)

        return delta_x

    @tf.function
    def _step(self, x: tf.Tensor, y_pred: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        """
        The optimization step to find the distance between the boundary and a given sample x.

        Notes
        -----
        To see more details about the optimization procedure for multi-class classifier, please
        refer to [https://arxiv.org/abs/1511.04599]

        Parameters
        ----------
        x
            The current sample used to compute the distance to the boundary of the model
        y_pred
            The one-hot labels predicted by the current model for the x sample

        Returns
        -------
        computation
            Boolean to determine if the optimization process should continue.
            True if the sample didn't change of class, False if not.
        loss_value
            The loss of the optimization procedure
        x_new
            The sample updated by the optimization procedure
        """
        y_pred = tf.argmax(y_pred, axis=1)

        with tf.GradientTape(watch_accessed_variables=False) as tape:
            tape.watch(x)
            y = self.model(x)

        y_computed = tf.argmax(y, axis=1)

        def update_grads():
            jac = tape.jacobian(y, x)

            indexes_all = tf.repeat(tf.expand_dims(tf.range(0, tf.shape(y)[1]), axis=0), tf.shape(y)[0], axis=0)
            indexes_class = tf.cast(tf.repeat(tf.expand_dims(y_pred, axis=1), tf.shape(y)[1], axis=1), dtype=tf.int32)
            indexes_other = tf.reshape(indexes_all[indexes_all != indexes_class], (-1, tf.shape(y)[1] - 1))

            delta_y = self.__delta_to_index(indexes_other, y_pred, y)
            delta_y = tf.abs(tf.reduce_mean(delta_y, axis=0))

            jac_delta = tf.reduce_mean(self.__delta_to_index(indexes_other, y_pred, jac), axis=0)

            jac_norm = tf.reshape(jac_delta, (tf.shape(jac_delta)[0], -1))
            jac_norm = tf.norm(jac_norm, axis=1)

            coeff = delta_y / jac_norm

            best_class = tf.argmin(coeff, axis=0)

            loss = (coeff / jac_norm)[best_class]

            x_new = x + loss * jac_delta[best_class]

            return loss, x_new

        computation = tf.reduce_any(y_computed == y_pred)

        top_k, _ = tf.math.top_k(tf.squeeze(y, axis=0), k=2)
        enough_close = tf.abs(top_k[0] - top_k[1]) > self.eps

        computation = tf.logical_and(computation, enough_close)

        loss_value, x_updated = tf.cond(computation, update_grads, lambda: (tf.constant(0.0, dtype=x.dtype), x))

        return computation, loss_value, x_updated

    def __compute_single_sample_score(self, x: tf.Tensor) -> tf.Tensor:
        """
        Computes the influence score (self-influence) for a single training sample.

        Parameters
        ----------
        x
            A tensor with a single training sample.

        Returns
        -------
        score
            The influence score of the sample.
        """
        x = tf.expand_dims(x, axis=0)
        y_pred = self.model(x)

        def body(index, x_current):
            computation, _, x_new = self._step(x_current, y_pred)
            return computation, index + 1, x_new

        _, _, x_adversarial = tf.while_loop(
            lambda cond, index, x_current: tf.logical_and(cond, index < self.step_nbr),
            lambda cond, index, x_current: body(index, x_current),
            [tf.constant(True), tf.constant(0, dtype=tf.int32), x])

        score = tf.norm(x - x_adversarial)

        return score

    def _compute_influence_value_from_batch(self, train_samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Computes the influence score (self-influence) for a single batch of training samples.

        Parameters
        ----------
        train_samples
            A tensor with a single batch of training sample.

        Returns
        -------
        influence_values
            The influence score of each sample in the batch train_samples.
        """
        scores = tf.map_fn(self.__compute_single_sample_score, train_samples[:-1][0], parallel_iterations=1)
        scores = - tf.expand_dims(scores, axis=1)

        return scores



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
First order Influence module implementing computations for all the different influence
related quantities: influence vector (delta of the weights after holding out the
sample and the original model's), Cook's distance (or influence values, a measure of
the model's reliance on the specific sample), both for individual points and whole
groups of data-points.

Disclaimer: This implements only a first order approximation of the influence function,
which does not take into account the pairwise interactions of data-points inside groups.
For a more precise (but much more computationally expensive) alternative, please refer
to the SecondOrderInfluenceCalculator module.
"""
import tensorflow as tf

from .base_group_influence import BaseGroupInfluenceCalculator

from ..common import InfluenceModel
from ..common import BaseInfluenceCalculator
from ..common import InverseHessianVectorProduct, IHVPCalculator

from ..types import Optional, Union, Tuple
from ..utils import assert_batched_dataset


class FirstOrderInfluenceCalculator(BaseInfluenceCalculator, BaseGroupInfluenceCalculator):
    """
    A class implementing the necessary methods to compute the different influence quantities
    using a first-order approximation. This makes it ideal for individual points and small
    groups of data, as it does so (relatively) efficiently.

    Notes
    -----
    For estimating the influence of large groups of data, please refer to the
    SecondOrderInfluenceCalculator class, which also takes into account the pairwise interactions
    between the points inside these groups.

    The methods currently implemented are available to evaluate one or a group of point(s):
    - Influence function vectors: the weights difference when removing points or groups of points
    - Influence values/Cook's distance: a measure of reliance of the model on the individual
      points or groups of points.

    For individual points, the following paper is used:
    [https://arxiv.org/abs/1703.04730](https://arxiv.org/abs/1703.04730).
    For groups of points, the following paper is used:
    [https://arxiv.org/abs/1905.13289](https://arxiv.org/abs/1905.13289)

    Parameters
    ----------
    model
        The TF2.X model implementing the InfluenceModel interface.
    dataset
        A batched TF dataset containing the training dataset over which we will estimate the
        inverse-hessian-vector product.
    ihvp_calculator
        Either a string containing the IHVP method ('exact' or 'cgd'), an IHVPCalculator
        object or an InverseHessianVectorProduct object.
    n_samples_for_hessian
        An integer indicating the amount of samples to take from the provided train dataset.
    shuffle_buffer_size
        An integer indicating the buffer size of the train dataset's shuffle operation -- when
        choosing the amount of samples for the hessian.
    normalize
        Implement "RelatIF: Identifying Explanatory Training Examples via Relative Influence"
        https://arxiv.org/pdf/2003.11630.pdf
        if True, compute the relative influence by normalizing the influence function.
    """

    def __init__(
            self,
            model: InfluenceModel,
            dataset: tf.data.Dataset,
            ihvp_calculator: Union[str, InverseHessianVectorProduct, IHVPCalculator] = 'exact',
            n_samples_for_hessian: Optional[int] = None,
            shuffle_buffer_size: Optional[int] = 10000,
            normalize=False
    ):
        super().__init__(
            model,
            dataset,
            ihvp_calculator,
            n_samples_for_hessian,
            shuffle_buffer_size
        )

        self.normalize = normalize

    @tf.function
    def _normalize_if_needed(self, v):
        """
        Normalize the input vector if the normalize property is True. If False, do nothing

        Parameters
        ----------
        v
            The vector to normalize of shape [Features_Space, Batch_Size]

        Returns
        -------
        v
            The normalized vector if the normalize property is True, otherwise the input vector
        """
        if self.normalize:
            v = v / tf.norm(v, axis=0, keepdims=True)
        return v

    @tf.function
    def _compute_influence_vector(self, train_samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Computes the influence vector (i.e. the delta of model's weights after a perturbation on the training
        dataset) for a single batch of training samples.

        Parameters
        ----------
        train_samples
            A tuple with the batch of training samples (with their labels).

        Returns
        -------
        influence_vector
            A tensor with the influence vector for each individual point. Shape will be (batch_size, nb_weights).
        """
        influence_vector = self.ihvp_calculator._compute_ihvp_single_batch(train_samples)  # pylint: disable=W0212
        influence_vector = self._normalize_if_needed(influence_vector)
        influence_vector = tf.transpose(influence_vector)
        return influence_vector

    @tf.function
    def _preprocess_samples(self, samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Preprocess a sample to evaluate

        Parameters
        ----------
        samples
            sample to evaluate

        Returns
        -------
        The preprocessed sample to evaluate
        """
        sample_evaluate_grads = self.model.batch_jacobian_tensor(samples)
        return sample_evaluate_grads

    def _estimate_individual_influence_values_from_batch(
            self,
            train_samples: Tuple[tf.Tensor, ...],
            samples_to_evaluate: Tuple[tf.Tensor, ...]
    ) -> tf.Tensor:
        """
        Estimate the (individual) influence scores of a single batch of samples with respect to
        a batch of samples belonging to the model's training dataset.

        Parameters
        ----------
        train_samples
            A single batch of training samples (and their target values).
        samples_to_evaluate
            A single batch of samples of which we wish to compute the influence of removing the training
            samples.

        Returns
        -------
        A tensor containing the individual influence scores.
        """
        return self._estimate_influence_value_from_influence_vector(
            self._preprocess_samples(samples_to_evaluate),
            self._compute_influence_vector(train_samples)
        )

    @tf.function
    def _estimate_influence_value_from_influence_vector(self, preproc_test_sample: tf.Tensor,
                                                        influence_vector: tf.Tensor) -> tf.Tensor:
        """
        Estimates the influence score of leaving out the influence vector corresponding to a given training
        data-point on a test sample that has already been pre-processed.

        Parameters
        ----------
        preproc_test_sample
            A single pre-processed test sample we wish to evaluate.
        influence_vector
            A single influence vector corresponding to a data-point from the training dataset.

        Returns
        -------
        influence_values
            A tensor with the resulting influence value.
        """
        influence_values = tf.matmul(preproc_test_sample, tf.transpose(influence_vector))
        return influence_values

    @tf.function
    def _compute_influence_value_from_batch(self, train_samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Computes the influence score (self-influence) for a single batch of training samples.

        Parameters
        ----------
        train_samples
            A tensor with a single batch of training sample.

        Returns
        -------
        influence_values
            The influence score of each sample in the batch train_samples.
        """
        batched_inf_vect = self._compute_influence_vector(train_samples)
        evaluate_vect = self._preprocess_samples(train_samples)
        influence_values = tf.reduce_sum(
            tf.math.multiply(evaluate_vect, batched_inf_vect), axis=1, keepdims=True)
        #TODO: improve IHVP to not compute 2 times the gradient
        return influence_values

    def compute_influence_vector_group(
            self,
            group: tf.data.Dataset
    ) -> tf.Tensor:
        """
        Computes the influence function vector -- an estimation of the weights difference when
        removing the points -- of the whole group of points.

        Parameters
        ----------
        group
            A batched TF dataset containing the group of points of which we wish to compute the
            influence of removal.

        Returns
        -------
        influence_group
            A tensor containing one vector for the whole group.
        """
        assert_batched_dataset(group)

        ihvp_ds = self.ihvp_calculator.compute_ihvp(group)
        reduced_ihvp = ihvp_ds.map(lambda x: tf.reduce_sum(x, axis=1, keepdims=True))
        reduced_ihvp = reduced_ihvp.reduce(tf.constant(0, dtype=ihvp_ds.element_spec.dtype), lambda x, y: x + y)

        reduced_ihvp = self._normalize_if_needed(reduced_ihvp)

        influence_group = tf.reshape(reduced_ihvp, (1, -1))

        return influence_group

    def estimate_influence_values_group(
            self,
            group_train: tf.data.Dataset,
            group_to_evaluate: Optional[tf.data.Dataset] = None
    ) -> tf.Tensor:
        """
        Computes Cook's distance of the whole group of points provided, giving measure of the
        influence that the group carries on the model's weights.

        The dataset_train contains the points we will be removing and dataset_to_evaluate,
        those with respect to whom we will be measuring the influence. As we will be performing
        the same operation in batches, we consider that each point from one dataset corresponds
        to one from the other. As such, both datasets must contain the same amount of points.
        In case the group_to_evaluate is not given, use by default the
        group_to_train: compute the self influence of the group.


        Parameters
        ----------
        group_train
            A batched TF dataset containing the group of points we wish to remove.
        group_to_evaluate
            A batched TF dataset containing the group of points with respect to whom we wish to
            measure the influence of removing the training points.

        Returns
        -------
        influence_values_group
            A tensor containing one influence value for the whole group.
        """
        if group_to_evaluate is None:
            # default to self influence
            group_to_evaluate = group_train

        ds_size = self.assert_compatible_datasets(group_train, group_to_evaluate)

        reduced_grads = tf.reduce_sum(tf.reshape(self.model.batch_jacobian(group_to_evaluate),
                                                 (ds_size, -1)), axis=0, keepdims=True)

        ihvp_ds = self.ihvp_calculator.compute_ihvp(group_train)
        reduced_ihvp = ihvp_ds.map(lambda x: tf.reduce_sum(x, axis=1, keepdims=True))
        reduced_ihvp = reduced_ihvp.reduce(tf.constant(0, dtype=ihvp_ds.element_spec.dtype), lambda x, y: x + y)

        reduced_ihvp = self._normalize_if_needed(reduced_ihvp)

        influence_values_group = tf.matmul(reduced_grads, reduced_ihvp)

        return influence_values_group



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module implementing factories for the different influence calculation techniques.
This will be useful for streamlining the benchmarks.
"""
from abc import abstractmethod

import numpy as np
import tensorflow as tf
from tensorflow.keras.losses import Loss, Reduction, CategoricalCrossentropy  # pylint: disable=E0611

from ..common import InfluenceModel, BaseInfluenceCalculator
from ..common import ExactIHVP, ConjugateGradientDescentIHVP, LissaIHVP
from ..common import ExactIHVPFactory, CGDIHVPFactory, LissaIHVPFactory

from ..influence import FirstOrderInfluenceCalculator, ArnoldiInfluenceCalculator
from ..rps import RepresenterPointLJE
from ..trac_in import TracIn
from ..rps import RepresenterPointL2
from ..boundary_based import WeightsBoundaryCalculator, SampleBoundaryCalculator

from ..types import Any, Union, Callable


class InfluenceCalculatorFactory:
    """
    An interface for factories generating instances of the different influence calculators.
    """

    @abstractmethod
    def build(self, training_dataset: tf.data.Dataset, model: tf.keras.Model,
              train_info: Any) -> BaseInfluenceCalculator:
        """
        Builds an instance of an influence calculator class following the provided model, training dataset
        and additional information.

        Parameters
        ----------
        training_dataset
            A TF dataset with the data on which the model was trained.
        model
            A TF model for which to compute the influence-related quantities.
        train_info
            An object providing additional information about the training procedure. For example, additional
            information is needed for computing influence values using TracIn.

        Returns
        -------
        The desired influence calculator instance.
        """
        raise NotImplementedError


class FirstOrderFactory(InfluenceCalculatorFactory):
    """
    A factory for creating instances of FirstOrderInfluenceCalculator objects.

    Attributes
    ----------
    ihvp_mode
        A string indicating whether the IHVPs should be computed using the 'exact' or the 'cgd' method.
    start_layer
        An integer for the target layer on which to compute the influence. By default, the last layer of
        the model is chosen.
    dataset_hessian_size
        An integer for the amount of samples that should go into the computation of the hessian matrix. By
        default, the entire training dataset is used.
    n_opt_iters
        An integer indicating how many iterations of the Conjugate Gradient Descent algorithm should be run
        before prematurely stopping the optimization.
    feature_extractor
        Either an integer for the last layer of the feature extractor, or an entire TF graph for computing the
        embeddings of the samples. Used if ihvp_mode == 'cgd'.
    loss_function
        Loss function to calculate influence (e.g. keras CategoricalCrossentropy). Make sure not to
        apply any reduction (Reduction.NONE), and specify correctly if the output is `from_logits`
        for example.
    """

    def __init__(self, ihvp_mode: str, start_layer=-1, dataset_hessian_size=-1, n_opt_iters=100,
                 feature_extractor: Union[int, tf.keras.Model] = -1,
                 loss_function: Callable = tf.keras.losses.CategoricalCrossentropy(
                     from_logits=True, reduction=Reduction.NONE)
                 ):
        self.start_layer = start_layer
        self.ihvp_mode = ihvp_mode
        self.n_opt_iters = n_opt_iters
        self.feature_extractor = feature_extractor
        self.dataset_hessian_size = dataset_hessian_size
        self.loss_function = loss_function
        assert self.ihvp_mode in ['exact', 'cgd', 'lissa']

    def build(self, training_dataset: tf.data.Dataset, model: tf.keras.Model,
              train_info: Any = None) -> FirstOrderInfluenceCalculator:
        """
        Builds an instance of the FirstOrderInfluenceCalculator class following the provided model and
        training dataset. No additional information is required.

        Parameters
        ----------
        training_dataset
            A TF dataset with the data on which the model was trained.
        model
            A TF model for which to compute the influence-related quantities.
        train_info
            In this case, None, as no additional information is required.

        Returns
        -------
        The desired FirstOrderInfluenceCalculator instance.
        """
        influence_model = InfluenceModel(model, start_layer=self.start_layer, loss_function=self.loss_function)

        if self.dataset_hessian_size is None or self.dataset_hessian_size < 0:
            dataset_hessian = training_dataset
        else:
            batch_size = training_dataset._batch_size.numpy()  # pylint: disable=W0212
            take_size = int(
                np.ceil(float(self.dataset_hessian_size) / batch_size)) * batch_size
            dataset_hessian = training_dataset.take(take_size)

        if self.ihvp_mode == 'exact':
            ihvp_calculator = ExactIHVP(influence_model, dataset_hessian)
        elif self.ihvp_mode == 'cgd':
            ihvp_calculator = ConjugateGradientDescentIHVP(influence_model, self.feature_extractor, dataset_hessian,
                                                           self.n_opt_iters)
        elif self.ihvp_mode == 'lissa':
            ihvp_calculator = LissaIHVP(influence_model, self.feature_extractor, dataset_hessian, self.n_opt_iters,
                                        damping=1e-4, scale=5.)
        else:
            raise ValueError("unknown ihvp calculator=" + self.ihvp_mode)

        influence_calculator = FirstOrderInfluenceCalculator(influence_model, training_dataset, ihvp_calculator,
                                                             n_samples_for_hessian=self.dataset_hessian_size)
        return influence_calculator


class RPSLJEFactory(InfluenceCalculatorFactory):
    """
    A factory for creating instances of representer point LJE objects.

    Attributes
    ----------
    ihvp_mode
        A string indicating whether the IHVPs should be computed using the 'exact' or the 'cgd' method.
    start_layer
        An integer for the target layer on which to compute the influence. By default, the last layer of
        the model is chosen.
    dataset_hessian_size
        An integer for the amount of samples that should go into the computation of the hessian matrix. By
        default, the entire training dataset is used.
    n_opt_iters
        An integer indicating how many iterations of the Conjugate Gradient Descent algorithm should be run
        before prematurely stopping the optimization.
    feature_extractor
        Either an integer for the last layer of the feature extractor, or an entire TF graph for computing the
        embeddings of the samples. Used if ihvp_mode == 'cgd'.
    loss_function
        Loss function to calculate influence (e.g. keras CategoricalCrossentropy). Make sure not to
        apply any reduction (Reduction.NONE), and specify correctly if the output is `from_logits`
        for example.
    """

    def __init__(self, ihvp_mode: str, start_layer=-1, dataset_hessian_size=-1, n_opt_iters=100,
                 feature_extractor: Union[int, tf.keras.Model] = -1,
                 loss_function: Callable = tf.keras.losses.CategoricalCrossentropy(
                     from_logits=True, reduction=Reduction.NONE)
                 ):
        self.start_layer = start_layer
        self.ihvp_mode = ihvp_mode
        self.n_opt_iters = n_opt_iters
        self.feature_extractor = feature_extractor
        self.dataset_hessian_size = dataset_hessian_size
        self.loss_function = loss_function
        assert self.ihvp_mode in ['exact', 'cgd', 'lissa']

    def build(self, training_dataset: tf.data.Dataset, model: tf.keras.Model,
              train_info: Any = None) -> RepresenterPointLJE:
        """
        Builds an instance of the RepresenterPointLJE class following the provided model and training dataset.
        No additional information is required in this case.

        Parameters
        ----------
        training_dataset
            A TF dataset with the data on which the model was trained.
        model
            A TF model for which to compute the influence-related quantities.
        train_info
            None in this case, as no additional information is required

        Returns
        -------
        The desired RepresenterPointLJE instance.
        """
        influence_model = InfluenceModel(model, start_layer=self.start_layer, loss_function=self.loss_function)

        if self.dataset_hessian_size is None or self.dataset_hessian_size < 0:
            dataset_hessian = training_dataset
        else:
            batch_size = training_dataset._batch_size.numpy()  # pylint: disable=W0212
            dataset_hessian = training_dataset.unbatch().take(self.dataset_hessian_size).batch(batch_size)

        if self.ihvp_mode == 'exact':
            ihvp_calculator_factory = ExactIHVPFactory()
        elif self.ihvp_mode == 'cgd':
            ihvp_calculator_factory = CGDIHVPFactory(self.feature_extractor, self.n_opt_iters)
        elif self.ihvp_mode == 'lissa':
            ihvp_calculator_factory = LissaIHVPFactory(self.feature_extractor, self.n_opt_iters, damping=1e-4, scale=5.)
        else:
            raise ValueError("unknown ihvp calculator=" + self.ihvp_mode)

        influence_calculator = RepresenterPointLJE(influence_model, dataset_hessian, ihvp_calculator_factory)
        return influence_calculator


class TracInFactory(InfluenceCalculatorFactory):
    """
    A factory for creating instances of TracIn objects.
    As it works by tracking the gradients along the training process, it also requires some
    training information to be able to compute influence values.

    Attributes
    ----------
    loss_function
        Loss function to calculate influence (e.g. keras CategoricalCrossentropy). Make sure not to
        apply any reduction (Reduction.NONE), and specify correctly if the output is `from_logits`
        for example.
    """

    def __init__(self, loss_function: Callable = CategoricalCrossentropy(from_logits=True, reduction=Reduction.NONE)):
        self.loss_function = loss_function

    def build(self, training_dataset: tf.data.Dataset, model: tf.keras.Model,
              train_info: Any) -> TracIn:
        """
        Builds an instance of the TracIn class following the provided model, training dataset
        and additional information.

        Parameters
        ----------
        training_dataset
            A TF dataset with the data on which the model was trained.
        model
            A TF model for which to compute the influence-related quantities.
        train_info
            A tuple with a list with the model's checkpoints on the first element and the corresponding
            list of learning rates on the second element.

        Returns
        -------
        The desired TracIn instance.
        """
        models = []
        for model_data in train_info[0]:
            influence_model = InfluenceModel(model_data, loss_function=self.loss_function)
            models.append(influence_model)
        influence_calculator = TracIn(models, train_info[1])
        return influence_calculator


class RPSL2Factory(InfluenceCalculatorFactory):
    """
    A factory for creating instances of RepresenterPointL2 objects.

    Attributes
    ----------
    loss_function
        The loss function with which the model was trained. This loss function MUST NOT be reduced.
    lambda_regularization
        The strength of the L2 regularization to add to the surrogate last layer.
    scaling_factor
        The Backtracking line-search's scaling factor for training the surrogate last layer. By default, this
        value is set to 0.1 and should typically converge quite easily.
    layer_index
        The index for the layer on which to compute the influence values.
    epochs
        An integer indicating for how long the surrogate last layer should be trained. By default, a value of
        100 is chosen.
    """

    def __init__(
            self,
            loss_function: Union[Callable[[tf.Tensor, tf.Tensor], tf.Tensor], Loss],
            lambda_regularization: float,
            scaling_factor: float = 0.1,
            layer_index=-2,
            epochs: int = 100
    ):
        self.loss_function = loss_function
        self.lambda_regularization = lambda_regularization
        self.scaling_factor = scaling_factor
        self.epochs = epochs
        self.layer_index = layer_index

    def build(self, training_dataset: tf.data.Dataset, model: tf.keras.Model,
              train_info: Any = None) -> RepresenterPointL2:
        """
        Builds an instance of the RepresenterPointL2 class following the provided model and training dataset.
        No additional information is required in this case.

        Parameters
        ----------
        training_dataset
            A TF dataset with the data on which the model was trained.
        model
            A TF model for which to compute the influence-related quantities.
        train_info
            None, as no additional information is required in this case.

        Returns
        -------
        The desired RepresenterPointL2 instance.
        """
        influence_calculator = RepresenterPointL2(model,
                                                  training_dataset,
                                                  self.loss_function,
                                                  self.lambda_regularization,
                                                  self.scaling_factor,
                                                  self.epochs,
                                                  self.layer_index)
        return influence_calculator


class WeightsBoundaryCalculatorFactory(InfluenceCalculatorFactory):
    """
    A factory for creating instances of WeightsBoundaryCalculator objects.

    Attributes
    ----------
    step_nbr
        The number of iterations to search the boundary for. By default, a value of 100 is chosen.
    norm_type
        The norm type of the distance between the weights to measure the distance to the boundary.
        By default, the L2 distance is chosen.
    """
    def __init__(self, step_nbr: int = 100, norm_type: int = 2):
        self.step_nbr = step_nbr
        self.norm_type = norm_type

    def build(self, training_dataset: tf.data.Dataset, model: tf.keras.Model,
              train_info: Any = None) -> WeightsBoundaryCalculator:
        """
        Builds an instance of the WeightsBoundaryCalculator class following the provided model and training dataset.
        No additional information is required in this case.

        Parameters
        ----------
        training_dataset
            A TF dataset with the data on which the model was trained.
        model
            A TF model for which to compute the influence-related quantities.
        train_info
            None, as no additional information is required in this case.

        Returns
        -------
        The desired WeightsBoundaryCalculator instance.
        """
        return WeightsBoundaryCalculator(model, self.step_nbr, self.norm_type)


class SampleBoundaryCalculatorFactory(InfluenceCalculatorFactory):
    """
    A factory for creating instances of SampleBoundaryCalculator objects.

    Attributes
    ----------
    step_nbr
        The number of iterations to search the boundary for. By default, a value of 100 is chosen.
    """

    def __init__(self, step_nbr: int = 100):
        self.step_nbr = step_nbr

    def build(self, training_dataset: tf.data.Dataset, model: tf.keras.Model,
              train_info: Any = None) -> SampleBoundaryCalculator:
        """
        Builds an instance of the SampleBoundaryCalculator class following the provided model and training dataset.
        No additional information is required in this case.

        Parameters
        ----------
        training_dataset
            A TF dataset with the data on which the model was trained.
        model
            A TF model for which to compute the influence-related quantities.
        train_info
            None, as no additional information is required in this case.

        Returns
        -------
        The desired SampleBoundaryCalculator instance.
        """
        return SampleBoundaryCalculator(model, self.step_nbr)


class ArnoldiCalculatorFactory(InfluenceCalculatorFactory):
    """
    A factory for creating instances of ArnoldiInfluenceCalculator objects.

    Attributes
    ----------
    subspace_dim
        The dimension of the Krylov subspace for the Arnoldi algorithm.
    force_hermitian
        A boolean indicating if we should force the projected matrix to be hermitian before the eigenvalue computation.
    k_largest_eig_vals
        An integer for the amount of top eigenvalues to keep for the influence estimations.
    start_layer
        An integer for the target layer on which to compute the influence. By default, the last layer of
        the model is chosen.
    dataset_hessian_size
        An integer for the amount of samples that should go into the computation of the hessian matrix. By
        default, the entire training dataset is used.
    dtype
        Numeric type for the Krylov basis (tf.float32 by default).
    """
    def __init__(
            self,
            subspace_dim: int,
            force_hermitian: bool,
            k_largest_eig_vals: int,
            start_layer: int = -1,
            dataset_hessian_size: int = -1,
            loss_function: Callable = CategoricalCrossentropy(from_logits=True, reduction=Reduction.NONE),
            dtype: tf.dtypes = tf.float32
    ):
        self.subspace_dim = subspace_dim
        self.force_hermitian = force_hermitian
        self.k_largest_eig_vals = k_largest_eig_vals
        self.start_layer = start_layer
        self.loss_function = loss_function
        self.dataset_hessian_size = dataset_hessian_size
        self.dtype = dtype

    def build(self, training_dataset: tf.data.Dataset, model: tf.keras.Model,
              train_info: Any = None) -> ArnoldiInfluenceCalculator:
        """
        Builds an instance of the ArnoldiInfluenceCalculator class following the provided model and training dataset.
        No additional information is required in this case.

        Parameters
        ----------
        training_dataset
            A TF dataset with the data on which the model was trained.
        model
            A TF model for which to compute the influence-related quantities.
        train_info
            None, as no additional information is required in this case.

        Returns
        -------
        The desired ArnoldiInfluenceCalculator instance.
        """
        influence_model = InfluenceModel(model, start_layer=self.start_layer, loss_function=self.loss_function)

        if self.dataset_hessian_size is None or self.dataset_hessian_size < 0:
            dataset_hessian = training_dataset
        else:
            batch_size = training_dataset._batch_size.numpy()  # pylint: disable=W0212
            take_size = int(np.ceil(float(self.dataset_hessian_size) / batch_size)) * batch_size
            dataset_hessian = training_dataset.take(take_size)

        return ArnoldiInfluenceCalculator(influence_model, dataset_hessian, self.subspace_dim, self.force_hermitian,
                                          self.k_largest_eig_vals, self.dtype)



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module implementing plotting functions for image-type data.
"""
from math import ceil

import tensorflow as tf
import numpy as np
from matplotlib import pyplot as plt

from ..types import Optional, Union


def _normalize(image: Union[tf.Tensor, np.ndarray]) -> np.ndarray:
    """
    Normalize an image to the range [0, 1].

    Parameters
    ----------
    image
        Image to prepare.

    Returns
    -------
    image
        Image ready to be used with matplotlib (in range[0, 1]).
    """
    image = np.array(image, np.float32)

    image -= image.min()
    image /= image.max()

    return image


def plot_most_influential_images(
        influential_images: tf.data.Dataset,
        cols: int = 5,
        img_size: float = 2.,
        save_path: Optional[str] = None
):
    """
    Plots the collection of images in a grid.

    Parameters
    ----------
    influential_images
        A dataset with the influential images and their influence scores.
    cols
        An integer indicating the amount of columns for the plot.
    img_size
        Size of each subplots (in inches), considering we keep aspect ratio.
    save_path
        An optional string specifying the path where to save the plot. If None, the figure
        will be displayed directly on screen.
    """
    # Materialize the contents of the dataset with the influential images
    images, influence_values = list(influential_images)

    # Calculate the amount of rows needed for this plot
    rows = ceil(len(images) / cols)
    # get width and height of our images
    l_width, l_height = images.shape[1:-1]

    # define the figure margin, width, height in inch
    margin = 0.3
    spacing = 0.3
    figwidth = cols * img_size + (cols - 1) * spacing + 2 * margin
    figheight = rows * img_size * l_height / l_width + (rows - 1) * spacing + 2 * margin

    left = margin / figwidth
    bottom = margin / figheight

    fig = plt.figure()
    fig.set_size_inches(figwidth, figheight)

    fig.subplots_adjust(
        left=left,
        bottom=bottom,
        right=1. - left,
        top=1. - bottom,
        wspace=spacing / img_size,
        hspace=spacing / img_size * l_width / l_height
    )

    if influence_values is not None:
        for i, (image, value) in enumerate(zip(images, influence_values)):
            plt.subplot(rows, cols, i + 1)
            plt.imshow(_normalize(image))
            plt.axis('off')
            plt.gca().set_title(f'inf_value = {value:.2f}')
    else:
        for i, image in enumerate(images):
            plt.subplot(rows, cols, i + 1)
            plt.imshow(_normalize(image))
            plt.axis('off')

    if save_path is None:
        plt.show()
    else:
        plt.savefig(save_path)
        plt.close()


def plot_datacentric_explanations(
        image: Union[tf.Tensor, np.ndarray],
        explanations: tf.data.Dataset,
        cols: int = 5,
        img_size: float = 2.,
        save_path: Optional[str] = None
):
    """
    Plots the most influential points for a test image, serving as a 'data-centric'
    explanation for why the model predicted a certain way.

    Parameters
    ----------
    image
        A tensor or numpy array with the image we wish to explain.
    explanations
        A dataset with the top-k most influential images for the provided image and their
        corresponding influence values.
    cols
        An integer indicating the amount of columns for the plot.
    img_size
        Size of each subplots (in inches), considering we keep aspect ratio.
    save_path
        An optional string specifying the path where to save the plot. If None, the figure
        will be displayed directly on screen.
    """
    # Calculate the amount of rows needed for this plot
    rows = ceil(len(explanations) / cols) + 1
    # get width and height of our images
    l_width, l_height = image.shape[:-1]

    # define the figure margin, width, height in inch
    margin = 0.3
    spacing = 0.3
    figwidth = cols * img_size + (cols - 1) * spacing + 2 * margin
    figheight = rows * img_size * l_height / l_width + (rows - 1) * spacing + 2 * margin

    left = margin / figwidth
    bottom = margin / figheight

    fig = plt.figure()
    fig.set_size_inches(figwidth, figheight)

    fig.subplots_adjust(
        left=left,
        bottom=bottom,
        right=1. - left,
        top=1. - bottom,
        wspace=spacing / img_size,
        hspace=spacing / img_size * l_width / l_height
    )

    plt.subplot(rows + 1, cols, 1)
    plt.imshow(_normalize(image))
    plt.axis('off')
    for i, (top_k_image, top_k_inf_value) in enumerate(zip(*explanations), start=1):
        plt.subplot(rows + 1, cols, i + cols)
        plt.imshow(_normalize(top_k_image))
        plt.axis('off')
        plt.gca().set_title(f'inf_value = {top_k_inf_value:.2f}')

    if save_path is None:
        plt.show()
    else:
        plt.savefig(save_path)
        plt.close()



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module defining some basic functionality for streamlining the process of benchmarking
the different influence calculation techniques. In particular, the main evaluation
task will be the detection of mislabeled examples (synthetically generated by adding
noise to labels in the training set) by looking at the most self-influential examples.
"""
from abc import abstractmethod
import os
import random
import json

import tensorflow as tf
import numpy as np
from tensorflow.keras.optimizers import Optimizer # pylint: disable=E0611

from .influence_factory import InfluenceCalculatorFactory
from ..types import Tuple, Dict, Any, Optional, List


class BaseTrainingProcedure:
    """
    A basic interface for neural network's learning procedure for streamlining batch
    training groups of models and facilitate the statistical test of the influence
    calculation's methods performances.
    """

    @abstractmethod
    def train(
            self,
            training_dataset: tf.data.Dataset,
            test_dataset: tf.data.Dataset,
            train_batch_size: int = 128,
            test_batch_size: int = 128,
            log_path: Optional[str] = None
    ) -> Tuple[float, float, tf.keras.Model, Any]:
        """
        Trains the model on the training dataset using a given size for the batches, and
        performs validation on the test dataset, eventually saving the model in the provided
        path (if not None).

        Parameters
        ----------
        training_dataset
            A TF dataset with all the data on which the model must be trained.
        test_dataset
            A TF dataset with all the data on which to validate the model's performance.
        train_batch_size
            An integer with the size of the batches on which to train the model.
        test_batch_size
            An integer with the size of the batches on which to perform the validation.
        log_path
            An (optional) string specifying where to save the model (if desired).

        Returns
        -------
        A tuple with: (the train accuracy, the test accuracy, the final model, the information from the model saver)
        """
        raise NotImplementedError


class MislabelingDetectorEvaluator:
    """
    A class implementing a benchmarking pipeline for influence calculators based on their capacity
    to point out the most self-influential examples as being the (voluntarily added) mislabeled
    examples in the training dataset.

    Notes
    -----
    As such, the experiments will consist on training a model using the specified procedure on a
    noisy dataset, computing the self-influence of each of the training dataset's points, sorting
    them by it and seeing how fast we can find all the mislabeled examples. In an industrial setting,
    the derivative of this curve close to zero will be paramount, as this means that we can find a
    considerable percentage of these samples by looking at a low percentage of the dataset (and thus,
    be done by a human operator).

    Parameters
    ----------
    training_dataset
            A TF dataset with all the data on which the model must be trained.
    test_dataset
        A TF dataset with all the data on which to validate the model's performance.
    training_procedure
        An object implementing the BaseTrainingProcedure interface and describing how the model
        must be trained.
    nb_classes
        An integer with the amount of classes of the classification problem.
    mislabeling_ratio
        A float with the ratio of noise to add to the training dataset's labels (ranging from 0. to 1.).
    train_batch_size
        An integer with the size of the batches on which to train the model.
    test_batch_size
        An integer with the size of the batches on which to perform the validation.
    config
        A dictionary with the configuration to save it alongside the results for traceability.
    """

    def __init__(
            self,
            training_dataset: tf.data.Dataset,
            test_dataset: tf.data.Dataset,
            training_procedure: BaseTrainingProcedure,
            nb_classes: int,
            mislabeling_ratio: float,
            train_batch_size: int = 128,
            test_batch_size: int = 128,
            influence_batch_size: Optional[int] = None,
            config: Optional[Dict] = None) -> None:
        assert 0. < mislabeling_ratio < 1.
        self.training_dataset = training_dataset
        self.train_batch_size = train_batch_size
        self.test_dataset = test_dataset
        self.test_batch_size = test_batch_size
        self.training_procedure = training_procedure
        self.nb_classes = nb_classes
        self.mislabeling_ratio = mislabeling_ratio
        if influence_batch_size is None:
            self.influence_batch_size = self.train_batch_size
        else:
            self.influence_batch_size = influence_batch_size

        if config is None:
            self.config = {}
        else:
            self.config = config

    def bench(
            self,
            influence_calculator_factories: Dict[str, InfluenceCalculatorFactory],
            nbr_of_evaluation: int,
            path_to_save: str,
            seed: int = 0,
            verbose: bool = True,
            use_tensorboard: bool = False
    ) -> Dict[str, Tuple[np.array, np.array, float]]:
        """
        Performs the whole benchmark for a group of influence calculator techniques and a number of
        evaluations for each of them (for statistical significance).

        Parameters
        ----------
        influence_calculator_factories
            A dictionary with the name of the influence calculator technique and a factory for creating
            instances for them.
        nbr_of_evaluation
            An integer with the amount of evaluations per method.
        path_to_save
            A string specifying the path to save the results.
        seed
            An integer for setting the seed on all the random number generators.
        verbose
            A boolean indicating whether progress is reported in stdout or not.
        use_tensorboard
            A boolean indicating if the results are to be progressively logged into tensorboard.

        Returns
        -------
        result
            A dictionary with the name of each method and its results.
        """
        result = {}
        for name, influence_calculator_factory in influence_calculator_factories.items():
            if verbose:
                print("starting to evaluate=" + str(name))

            curves, mean_curve, roc = self.evaluate(influence_calculator_factory, nbr_of_evaluation, seed, verbose,
                                                    path_to_save, use_tensorboard, name)

            result[name] = (curves, mean_curve, roc)

            if verbose:
                print(name + " | mean roc=" + str(roc))

        return result

    def evaluate(
            self,
            influence_factory: InfluenceCalculatorFactory,
            nbr_of_evaluation: int,
            seed: int = 0,
            verbose: bool = True,
            path_to_save: Optional[str] = None,
            use_tensorboard: bool = False,
            method_name: Optional[str] = None
    ) -> Tuple[np.array, np.array, float]:
        """
        Performs one benchmark evaluation over one influence calculator technique the specified number
        of times (for statistical significance).

        Parameters
        ----------
        influence_factory
            A factory for instantiating objects for a given influence calculator technique.
        nbr_of_evaluation
            An integer with the amount of evaluations per method.
        seed
            An integer for setting the seed on all the random number generators.
        verbose
            A boolean indicating whether progress is reported in stdout or not.
        path_to_save
            A string specifying the path to save the results.
        use_tensorboard
            A boolean indicating if the results are to be progressively logged into tensorboard.
        method_name
            An optional string with the experience's name.

        Returns
        -------
        curves, mean_curve, roc
            A tuple with the experience's results: (each of the individual curves, the mean curve, the ROC)
        """
        curves = []

        if use_tensorboard and (path_to_save is None):
            path_to_save = "./"

        if path_to_save is not None:
            dirname = path_to_save + "/" + method_name
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            with open(dirname + "/config.json", 'w', encoding="utf-8") as fp:
                json.dump(self.config, fp, indent=4)

        if method_name is None:
            method_name = 'experiment'

        for index in range(nbr_of_evaluation):

            if use_tensorboard:
                experiment_name = method_name + "_" + str(index)

                file_writer = tf.summary.create_file_writer(path_to_save + "/" + method_name + "/seed" + str(index),
                                                            filename_suffix=experiment_name)
                tf_writer = file_writer.as_default()

            tf.keras.backend.clear_session()
            self.set_seed(seed + index)

            noisy_training_dataset, noisy_label_indexes = self.build_noisy_training_dataset()
            noisy_label_indexes = noisy_label_indexes[0]

            acc_train, acc_test, model, data_train = self.training_procedure.train(
                noisy_training_dataset,
                self.test_dataset,
                self.train_batch_size,
                self.test_batch_size,
                log_path=None if path_to_save is None else path_to_save + "/" + method_name + "/seed" + str(index))

            influence_calculator = influence_factory.build(
                noisy_training_dataset.shuffle(1000).batch(self.influence_batch_size), model, data_train)

            influences_values = influence_calculator._compute_influence_values(  # pylint: disable=W0212
                noisy_training_dataset.batch(self.influence_batch_size))

            # compute curve and indexes
            sorted_influences_indexes = np.argsort(-np.squeeze(influences_values))

            sorted_curve = self.__compute_curve(sorted_influences_indexes, noisy_label_indexes)
            curves.append(sorted_curve)

            roc = self._compute_roc(sorted_curve)
            if verbose:
                print("seed nbr=" + str(index) + " | acc train=" + str(acc_train) + " | acc test=" + str(
                    acc_test) + " | roc=" + str(roc))

            if use_tensorboard:
                with tf_writer:
                    tf.summary.scalar("roc_value", roc, index)
                    self.plot_tensorboard_roc(sorted_curve, "roc_curve")

            if path_to_save is not None:
                curves_, mean_curve_, roc_ = self.__build(curves)
                self.__save(path_to_save + "/" + method_name + "/data.npy", curves_, mean_curve_, roc_)

        curves, mean_curve, roc = self.__build(curves)

        if use_tensorboard:
            file_writer = tf.summary.create_file_writer(path_to_save + "/synthesis/" + method_name + "/")
            with file_writer.as_default():
                tf.summary.scalar("roc_mean", roc, 0)
                tf.summary.scalar("roc_mean", roc, 1)
                self.plot_tensorboard_roc(mean_curve, "roc_curve_mean")

        return curves, mean_curve, roc

    @staticmethod
    def plot_tensorboard_roc(curve: np.ndarray, experiment_name: str):
        """
        Plots a mislabeled samples detection ROC curve on tensorboard.

        Parameters
        ----------
        curve
            A numpy array with the experiment's curve
        experiment_name
            A string with the experiment's name
        """
        for i, c in enumerate(curve):
            tf.summary.scalar(experiment_name, c, i)

    def __build(self, curves: List[np.ndarray]) -> Tuple[np.ndarray, np.ndarray, float]:
        """
        Formats the curves, computes the mean curve and the ROC.

        Parameters
        ----------
        curves
            A list with the experiment's curves mislabeled sample detection curves.
        Returns
        -------
        curves, mean_curve, roc
            A tuple with: (curves in numpy array format, the mean curve, the roc value)
        """
        curves = np.asarray(curves)
        mean_curve = np.mean(curves, axis=0)
        roc = self._compute_roc(mean_curve)

        return curves, mean_curve, roc

    @staticmethod
    def _compute_roc(curve: np.array) -> float:
        """
        Computes the ROC value of the curve.

        Parameters
        ----------
        curve
            A numpy array with a curve.

        Returns
        -------
        roc
            The roc value.
        """
        roc = np.mean(curve)
        return roc

    @staticmethod
    def set_seed(seed: int):
        """
        Sets all the random seeds on TensorFlow, numpy and python for traceability.

        Parameters
        ----------
        seed
            An integer with the seed value.
        """
        tf.random.set_seed(seed)
        np.random.seed(seed)
        random.seed(seed)

    @staticmethod
    def __compute_curve(sorted_influences_indexes: np.ndarray, noisy_label_indexes: np.ndarray) -> np.array:
        """
        Computes the mislabeled sample detection curve using the indices of the samples sorted
        by their self-influence and the ground-truth indices of the target points.

        Parameters
        ----------
        sorted_influences_indexes
            A numpy array with the sample's indices sorted by their self-influence.
        noisy_label_indexes
            A numpy array with the dataset's mislabeled examples' ground-truth.

        Returns
        -------
        curve
            A numpy array with the detection curve as we progressively scan the dataset.
        """
        index = np.in1d(sorted_influences_indexes, noisy_label_indexes)
        index = tf.cast(index, np.int32)
        curve = np.cumsum(index)
        if curve[-1] != 0:
            curve = curve / curve[-1]

        return curve

    def build_noisy_training_dataset(self) -> Tuple[tf.data.Dataset, np.array]:
        """
        Generates a noisy version of the object's own dataset. In particular, it will include noise
        in the label information (i.e. the label will be switched at random). More noise will
        effectively impact the model's capacity to predict correctly on the test set, as it will
        need to learn to use spurious correlations to attain the 100% accuracy on the training
        dataset.

        Returns
        -------
        noisy_dataset, noise_indexes
            A tuple with the noisy dataset and a numpy array with the flipped labels (used for validation
            during the evaluation).
        """
        dataset_size = tf.data.experimental.cardinality(self.training_dataset)
        noise_mask = np.random.uniform(size=(dataset_size,)) > self.mislabeling_ratio
        noise_mask_dataset = tf.data.Dataset.from_tensor_slices(noise_mask)
        noisy_dataset = tf.data.Dataset.zip((self.training_dataset, noise_mask_dataset))

        def noise_map(z, y_mask):
            """
            Flips a single sample's labels following the mask.
            """
            (x, y) = z
            y_noise = tf.random.uniform(shape=(1,)) * tf.cast((tf.shape(y)[-1] - 1), dtype=tf.float32)
            y_noise = tf.cond(y_noise > tf.cast(tf.argmax(y), dtype=tf.float32), lambda: y_noise + 1, lambda: y_noise)
            y_noise = tf.cast(y_noise, dtype=tf.int32)
            y_noise = tf.cast(tf.squeeze(tf.one_hot(y_noise, tf.shape(y)[-1]), axis=0), dtype=y.dtype)
            y = tf.where(y_mask, y, y_noise)
            return x, y

        noisy_dataset = noisy_dataset.map(noise_map)

        noise_indexes = np.where(np.logical_not(noise_mask))
        return noisy_dataset, noise_indexes

    @staticmethod
    def __save(path_to_save: str, curves: np.array, mean_curve: np.array, roc: float) -> None:
        """
        Saves an evaluation's results to the disk.

        Parameters
        ----------
        path_to_save
            A string with the path into which to save the results.
        curves
            A numpy array with the different detection curves.
        mean_curve
            A numpy array with the mean detection curve.
        roc
            A float with the ROC value.
        """
        dirname = os.path.dirname(path_to_save)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        np.save(path_to_save, np.array((curves, mean_curve, roc), dtype=object), allow_pickle=True)


class ModelsSaver(tf.keras.callbacks.Callback):
    """
    A simple class to save models after optimizer updates. It will prove itself useful for tracing the
    different information to be able to use the TracIn method.

    Parameters
    ----------
    epochs_to_save
        A list of integers indicating on which epochs to save a model's checkpoint.
    optimizer
        The model's optimizer.
    saving_path
        An (optional) string for saving the results on the disk.
    """

    def __init__(self, epochs_to_save: List[int], optimizer: Optimizer, saving_path: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)
        self.epochs_to_save = epochs_to_save
        self.optimizer = optimizer

        self.models = []
        self.learning_rates = []

        if saving_path is not None and not os.path.exists(saving_path):
            os.mkdir(saving_path)
        self.saving_path = saving_path

    def on_epoch_end(self, epoch: int, logs: Optional[Dict] = None) -> None:
        """
        Save the relevant training information (model, learning rate, save to disk if desired)
        after a training epoch.

        Parameters
        ----------
        epoch
            An integer with the current epoch. If it is in the list of epochs after which to
            save the training information, do so.
        logs
            Dict, metric results for this training epoch, and for the validation epoch if validation
            is performed. Validation result keys are prefixed with val_. For training epoch, the
            values of the Model's metrics are returned.
        """
        if epoch in self.epochs_to_save:
            epoch_model = tf.keras.models.clone_model(self.model)
            epoch_model.build(self.model.input_shape)
            epoch_model.set_weights(self.model.get_weights())

            epoch_lr = self.optimizer.lr
            self.models.append(epoch_model)
            self.learning_rates.append(epoch_lr.numpy())

            if self.saving_path is not None:
                tf.data.experimental.save(f"{self.saving_path}/model_ep_{epoch:.6d}")
                np.save(f"{self.saving_path}/learning_rates", np.array(self.learning_rates), allow_pickle=True)
                with open(f"{self.saving_path}/logs.json", "w", encoding='utf8') as f:
                    json.dump(logs, f)



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module implementing the TracIn method originally presented in
https://arxiv.org/pdf/2002.08484.pdf
It computes the influence of each training point by tracing the gradients during the
training phase. In practice, we will use the model at different checkpoints, achieving
a more efficient estimation at the cost of a little precision.
"""
import tensorflow as tf

from ..common import InfluenceModel, BaseInfluenceCalculator
from ..types import Union, List, Tuple


class TracIn(BaseInfluenceCalculator):
    """
    A class implementing an influence score based on TracIn method proposed in
    [https://arxiv.org/pdf/2002.08484.pdf](https://arxiv.org/pdf/2002.08484.pdf)

    Notes
    -----
    This method traces the gradients of each of the training data-points as a means to
    compute their influence in the model. In practice, instead of saving the gradients
    at each epoch, the model's checkpoints are used, with the checkpoint frequency
    allowing for a trade-off between the precision of the computed influence and its
    computational cost.

    Parameters
    ----------
    models
        A list of TF2.X models implementing the InfluenceModel interface at different steps (epochs)
        of the training
    learning_rates
        Learning rate or list of learning rates used during the training.
        If learning_rates is a list, it should have the same size as the amount of models
    """
    def __init__(self, models: List[InfluenceModel], learning_rates: Union[float, List[float]]):
        self.models = models

        if isinstance(learning_rates, List):
            assert len(models) == len(learning_rates)
            self.learning_rates = learning_rates
        else:
            self.learning_rates = [learning_rates for _ in range(len(models))]

    def _compute_influence_vector(self, train_samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Compute an equivalent of the influence vector for a sample of training points.

        Disclaimer: this vector is not an estimation of the difference between the actual
        model and the perturbed model without the samples (like it is the case with what is
        calculated using deel.influenciae.influence).

        Parameters
        ----------
        train_samples
            A tensor with a group of training samples of which we wish to compute the influence.

        Returns
        -------
        influence_vectors
            A tensor with the influence for each sample.
        """
        influence_vectors = []
        for model, lr in zip(self.models, self.learning_rates):
            g_train = model.batch_jacobian_tensor(train_samples)
            influence_vectors.append(g_train * tf.cast(tf.sqrt(lr), g_train.dtype))
        influence_vectors = tf.concat(influence_vectors, axis=1)

        return influence_vectors

    def _preprocess_samples(self, samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Pre-process a sample to facilitate evaluation afterwards. In this case, it amounts to transforming
        it into it's "influence vector".

        Parameters
        ----------
        samples
            A tensor with the group of samples we wish to evaluate.
        Returns
        -------
        evaluate_vect
            A tensor with the pre-processed samples.
        """
        evaluate_vect = self._compute_influence_vector(samples)
        return evaluate_vect

    def _estimate_individual_influence_values_from_batch(
            self,
            train_samples: Tuple[tf.Tensor, ...],
            samples_to_evaluate: Tuple[tf.Tensor, ...]
    ) -> tf.Tensor:
        """
        Estimate the (individual) influence scores of a single batch of samples with respect to
        a batch of samples belonging to the model's training dataset.

        Parameters
        ----------
        train_samples
            A single batch of training samples (and their target values).
        samples_to_evaluate
            A single batch of samples of which we wish to compute the influence of removing the training
            samples.

        Returns
        -------
        A tensor containing the individual influence scores.
        """
        return self._estimate_influence_value_from_influence_vector(
            self._preprocess_samples(samples_to_evaluate),
            self._compute_influence_vector(train_samples)
        )

    def _estimate_influence_value_from_influence_vector(
            self,
            preproc_test_sample: tf.Tensor,
            influence_vector: tf.Tensor
    ) -> tf.Tensor:
        """
        Compute the influence score of a (pre-processed) sample and an "influence vector" from a training
        data-point

        Parameters
        ----------
        preproc_test_sample
            A tensor with a (pre-processed) test sample
        influence_vector
            A tensor with an "influence vector" calculated using a training point

        Returns
        -------
        influence_values
            A tensor with the influence scores
        """
        influence_values = tf.matmul(preproc_test_sample, tf.transpose(influence_vector))
        return influence_values

    def _compute_influence_value_from_batch(self, train_samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Compute the influence score for a training sample

        Parameters
        ----------
        train_samples
            Training sample
        Returns
        -------
        The influence score
        """
        influence_vector = self._compute_influence_vector(train_samples)
        influence_values = tf.reduce_sum(influence_vector * influence_vector, axis=1, keepdims=True)
        return influence_values



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module implementing the necessary functions for computing influence functions on
groups of data -- i.e. performing approximate hold-one-out of groups of data-points
as a whole instead of individually. The theoretical validity of this implementation
was discussed in https://arxiv.org/abs/1905.13289 , with positive results.

We propose two classes that implement this interface: the FirstOrderInfluenceCalculator
(using the technique proposed in https://arxiv.org/abs/1905.13289) and the
SecondOrderInfluenceCalculator, that is only available for calculating groups'
influence and does so through the method presented in
https://arxiv.org/abs/1911.00418
"""

from abc import abstractmethod
import tensorflow as tf

from ..common import InfluenceModel
from ..common import InverseHessianVectorProduct, IHVPCalculator, ExactIHVP

from ..utils import dataset_size
from ..types import Optional, Union


class BaseGroupInfluenceCalculator:
    """
    A base class for objects that calculate the different quantities related to influence
    functions for whole groups of data-points.

    Notes
    -----
    The methods currently implemented are available to evaluate groups of points:
    - Influence function vectors: the weights difference when removing entire groups of points.
    - Influence values/Cook's distance: a measure of reliance of the model on the entire group.

    Parameters
    ----------
    model
        The TF2.X model implementing the InfluenceModel interface.
    dataset
        A batched TF dataset containing the training dataset over which we will estimate the
        inverse-hessian-vector product.
    ihvp_calculator
        Either a string containing the IHVP method ('exact' or 'cgd'), an IHVPCalculator
        object or an InverseHessianVectorProduct object.
    n_samples_for_hessian
        An integer indicating the amount of samples to take from the provided train dataset.
    shuffle_buffer_size
        An integer indicating the buffer size of the train dataset's shuffle operation -- when
        choosing the amount of samples for the hessian.
    """
    def __init__(
            self,
            model: InfluenceModel,
            dataset: tf.data.Dataset,
            ihvp_calculator: Union[str, InverseHessianVectorProduct, IHVPCalculator] = ExactIHVP,
            n_samples_for_hessian: Optional[int] = None,
            shuffle_buffer_size: Optional[int] = 10000
    ):
        self.model = model

        if n_samples_for_hessian is None:
            dataset_to_estimate_hessian = dataset
        else:
            dataset_to_estimate_hessian = dataset.unbatch().shuffle(shuffle_buffer_size)\
                .take(n_samples_for_hessian).batch(dataset._batch_size)

        self.train_set = dataset_to_estimate_hessian

        # load ivhp calculator from str, IHVPcalculator enum or InverseHessianVectorProduct object
        if isinstance(ihvp_calculator, str):
            self.ihvp_calculator = IHVPCalculator.from_string(ihvp_calculator).value(self.model, self.train_set) if \
                ihvp_calculator == 'exact' else \
                IHVPCalculator.from_string(ihvp_calculator).value(self.model,
                                                                  self.model.start_layer,
                                                                  self.train_set)
        elif isinstance(ihvp_calculator, IHVPCalculator):
            self.ihvp_calculator = ihvp_calculator.value(self.model, self.train_set)
        elif isinstance(ihvp_calculator, InverseHessianVectorProduct):
            self.ihvp_calculator = ihvp_calculator

    @abstractmethod
    def compute_influence_vector_group(
            self,
            group: tf.data.Dataset
    ) -> tf.Tensor:
        """
        Computes the influence function vector -- an estimation of the weights difference when
        removing the points -- of the whole group of points.

        Parameters
        ----------
        group
            A batched TF dataset containing the group of points of which we wish to compute the
            influence of removal.

        Returns
        -------
        influence_group
            A tensor containing one vector for the whole group.
        """
        raise NotImplementedError()

    @abstractmethod
    def estimate_influence_values_group(
            self,
            group_train: tf.data.Dataset,
            group_to_evaluate: Optional[tf.data.Dataset] = None
    ) -> tf.Tensor:
        """
        Computes Cook's distance of the whole group of points provided, giving measure of the
        influence that the group carries on the model's weights.

        The dataset_train contains the points we will be removing and dataset_to_evaluate,
        those with respect to whom we will be measuring the influence. As we will be performing
        the same operation in batches, we consider that each point from one dataset corresponds
        to one from the other. As such, both datasets must contain the same amount of points.
        In case the group_to_evaluate is not given, use by default the
        group_to_train: compute the self influence of the group.


        Parameters
        ----------
        group_train
            A batched TF dataset containing the group of points we wish to remove.
        group_to_evaluate
            A batched TF dataset containing the group of points with respect to whom we wish to
            measure the influence of removing the training points.

        Returns
        -------
        influence_values_group
            A tensor containing one influence value for the whole group.
        """
        raise NotImplementedError()

    @staticmethod
    def assert_compatible_datasets(dataset_a: tf.data.Dataset, dataset_b: tf.data.Dataset) -> int:
        """
        Assert that the datasets are compatible: that they contain the same number of points. Else,
        throw an error.

        Parameters
        ----------
        dataset_a
            First batched tensorflow dataset to check.
        dataset_b
            Second batched tensorflow dataset to check.

        Returns
        -------
        size
            The size of the dataset.
        """
        size_a = dataset_size(dataset_a)
        size_b = dataset_size(dataset_b)

        if size_a != size_b:
            raise ValueError("The amount of points in the train and evaluation groups must match.")

        return size_a



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Influenciae
-------

The goal of Influenciae is to provide a simple interface to the various influence functions
techniques
"""

__version__ = '0.3.0'

from . import influence
from . import common
from . import rps
from . import trac_in
from . import benchmark
from . import plots



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Influence calculator module implementing the approximations in ["Scaling up Influence
Functions"](https://arxiv.org/pdf/2112.03052.pdf). The Arnoldi algorithm effectively
reduces the dimension of the problem of computing IHVPs and allows for the calculation
of influence values on big neural network models.
"""
import tensorflow as tf

from ..common import InfluenceModel, BaseInfluenceCalculator, ForwardOverBackwardHVP
from ..types import Tuple


class ArnoldiInfluenceCalculator(BaseInfluenceCalculator):
    """
    A class implementing an influence score based on reducing the dimension of the problem
    of computing IHVPs through the Arnoldi algorithm as per https://arxiv.org/pdf/2112.03052.pdf
    This allows this calculator to be used on models with a considerable amount of weights
    in a time-efficient manner. The influence score being calculated is theoretically the
    same as the rest of the calculators in the `influence` sub-package.

    Notes
    -----
    This influence calculator applies several approximations to allow it to scale in the amount of model weights
    and in compute time. In [https://arxiv.org/pdf/2112.03052.pdf], the authors compute the influence for
    transformer models with tens of millions of parameters.

    Parameters
    ----------
    model
        The TF2.X model implementing the InfluenceModel interface.
    train_dataset
        A batched TF dataset with the points with which the model was trained.
    subspace_dim
        The dimension of the Krylov subspace for the Arnoldi algorithm.
    force_hermitian
        A boolean indicating if we should force the projected matrix to be hermitian before the eigenvalue computation.
    k_largest_eig_vals
        An integer for the amount of top eigenvalues to keep for the influence estimations.
    dtype
        Numeric type for the Krylov basis (tf.float32 by default).
    """
    def __init__(
            self,
            model: InfluenceModel,
            train_dataset: tf.data.Dataset,
            subspace_dim: int,
            force_hermitian: bool,
            k_largest_eig_vals: int,
            dtype: tf.dtypes = tf.float32
    ):
        self.subspace_dim = subspace_dim
        self.force_hermitian = force_hermitian
        self.k_largest_eig_vals = k_largest_eig_vals
        self.model = model
        self.hvp_calculator = ForwardOverBackwardHVP(model, train_dataset)
        self.dtype = dtype

        self.eig_vals, self.G = self.arnoldi(self.model.nb_params)

    def arnoldi(self, dim: int) -> Tuple[tf.Tensor, tf.Tensor]:
        """
        Builds the projection of the inverse of the hessian on the Krylov subspaces.

        Parameters
        ----------
        dim
            The dimension of the basis

        Returns
        -------
        eig_vals
            The eigen values of the projection
        G
            The projection matrix
        """
        v = tf.random.normal((dim,), dtype=self.dtype)
        A, W = self._build_orthogonal_basis(v)
        eig_vals, G = self._distill(A, W)

        return eig_vals, G

    @tf.function
    def __build_orthogonal_basis_iter(
            self,
            W: tf.Tensor,
            A: tf.Tensor,
            index: int
    ) -> Tuple[tf.Tensor, tf.Tensor, int]:
        """
        Builds the new vector of the Krylov's basis and computes the projection of the hessian for this vector.

        Parameters
        ----------
        W
            The (index+1) first vectors of the Krylov's basis
        A
            The "index" first vectors of the projection of the hessian on the Krylov's basis
        index
            The current index of the Krylov's basis

        Returns
        -------
        W
            The (index+2) first vectors of the Krylov's basis
        A
            The (index+1) first vectors of the projection of the hessian on the Krylov's basis
        index
            The current index of the Krylov's basis
        """
        w_next = self.hvp_calculator(W[index])
        w_next = tf.squeeze(w_next, axis=1)
        size = index + 1
        A_next_line = tf.reduce_sum(W[:size] * tf.repeat(tf.expand_dims(w_next, axis=0), size, axis=0), axis=1)
        WA_product = tf.reduce_sum(W[:size] * tf.repeat(tf.expand_dims(A_next_line, axis=1), tf.shape(W)[1], axis=1),
                                   axis=0)
        w_next = w_next - WA_product

        w_next_norm = tf.norm(w_next)

        padding_size = tf.shape(A)[1] - tf.shape(A_next_line)[0] - 1
        A_next_line = tf.concat(
            [A_next_line, tf.expand_dims(w_next_norm, axis=0), tf.zeros((padding_size,), self.dtype)],
            axis=0)

        W = tf.concat([W[:size], tf.expand_dims(w_next, axis=0) / w_next_norm,
                       tf.zeros((tf.shape(W)[0] - size - 1, tf.shape(W)[1]), self.dtype)], axis=0)

        A = tf.concat(
            [A[:index], tf.expand_dims(A_next_line, axis=0),
             tf.zeros((tf.shape(A)[0] - index - 1, tf.shape(A)[1]), self.dtype)],
            axis=0)

        return W, A, index + 1

    def _build_orthogonal_basis(self, v: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
        """
        Build orthonormal basis for the Krylov subspaces with the first vector of the basis v.
        Project the hessian on the Krylov subspaces.

        Parameters
        ----------
        v
            The first vector of the Krylov basis

        Returns
        -------
        A
            The projection of the hessian on the Krylov basis
        W
            The Krylov basis
        """
        W0 = tf.concat(
            [tf.expand_dims(v / tf.norm(v), axis=0), tf.zeros((self.subspace_dim, tf.shape(v)[0]), self.dtype)],
            axis=0)
        A0 = tf.zeros((self.subspace_dim, self.subspace_dim + 1), dtype=v.dtype)

        W, A, _ = tf.while_loop(lambda W, A, index: index < self.subspace_dim, self.__build_orthogonal_basis_iter,
                                [W0, A0, tf.constant(0, dtype=tf.int32)],
                                parallel_iterations=1)

        return A, W

    def _distill(self, A: tf.Tensor, W: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
        """
        Inverse the projection by performing the following operations:

            Hessian = W^T * A * W

            A = V^T * I * V
            G^T = W^T * V^T
            G = V W

            => Hessian = G^T * I * G

        Parameters
        ----------
        A
            The projection of the hessian on the Krylov basis
        W
            The Krylov basis

        Returns
        -------
        eig_vals
            The eigen values of the projection
        G
            The projection matrix
        """
        A = A[:, :-1]
        W = W[:-1, :]

        if self.force_hermitian:
            maindiag = tf.linalg.diag_part(A, k=0)
            superdiag = tf.linalg.diag_part(A, k=1)
            subdiag = tf.linalg.diag_part(A, k=-1)

            superdiag = (superdiag + subdiag) / 2.0

            with tf.device('cpu'):
                eig_vals, eig_vectors = tf.linalg.eigh_tridiagonal(maindiag, superdiag, eigvals_only=False)
        else:
            eig_vals, eig_vectors = tf.linalg.eig(A)

        _, idx = tf.math.top_k(- tf.abs(eig_vals), k=self.k_largest_eig_vals)
        eig_vals = tf.gather(eig_vals, idx, axis=-1)
        eig_vectors = tf.gather(eig_vectors, idx, axis=-1)

        G = tf.matmul(eig_vectors, tf.cast(W, dtype=eig_vectors.dtype), transpose_a=True)

        return eig_vals, G

    def _compute_influence_vector(self, train_samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Compute an equivalent of the influence vector for a sample of training points.

        Parameters
        ----------
        train_samples
            A tensor with a group of training samples of which we wish to compute the influence.

        Returns
        -------
        influence_vectors
            A tensor with the influence for each sample.
        """
        g_train = self.model.batch_jacobian_tensor(train_samples)

        influence_vectors = tf.matmul(tf.cast(g_train, dtype=self.G.dtype), self.G, transpose_b=True) / self.eig_vals

        return influence_vectors

    def _preprocess_samples(self, samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Pre-process a sample to facilitate evaluation afterwards. In this case, it amounts to transforming
        it into it's "influence vector".

        Parameters
        ----------
        samples
            A tensor with the group of samples we wish to evaluate.

        Returns
        -------
        evaluate_vect
            A tensor with the pre-processed samples.
        """
        g_sample = self.model.batch_jacobian_tensor(samples)

        evaluate_vect = tf.matmul(tf.cast(g_sample, self.G.dtype), self.G, transpose_b=True)

        return evaluate_vect

    def _estimate_influence_value_from_influence_vector(
            self,
            preproc_test_sample: tf.Tensor,
            influence_vector: tf.Tensor
    ) -> tf.Tensor:
        """
        Compute the influence score of a (pre-processed) sample and an "influence vector" from a training
        data-point

        Parameters
        ----------
        preproc_test_sample
            A tensor with a (pre-processed) test sample
        influence_vector
            A tensor with an "influence vector" calculated using a training point

        Returns
        -------
        influence_values
            A tensor with the influence scores
        """
        influence_values = tf.matmul(preproc_test_sample, tf.transpose(influence_vector))
        influence_values = tf.math.real(influence_values)

        return influence_values

    def _compute_influence_value_from_batch(self, train_samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Compute the influence score for a training sample

        Parameters
        ----------
        train_samples
            Training sample

        Returns
        -------
        The influence score
        """
        g_train = self.model.batch_jacobian_tensor(train_samples)
        influence_vectors = tf.matmul(tf.cast(g_train, dtype=self.G.dtype), self.G, transpose_b=True)

        influence_values = tf.reduce_sum(influence_vectors * influence_vectors / self.eig_vals, axis=1, keepdims=True)
        influence_values = tf.math.real(influence_values)

        return influence_values

    def _estimate_individual_influence_values_from_batch(
            self,
            train_samples: Tuple[tf.Tensor, ...],
            samples_to_evaluate: Tuple[tf.Tensor, ...]
    ) -> tf.Tensor:
        """
        Estimate the (individual) influence scores of a single batch of samples with respect to
        a batch of samples belonging to the model's training dataset.

        Parameters
        ----------
        train_samples
            A single batch of training samples (and their target values).
        samples_to_evaluate
            A single batch of samples of which we wish to compute the influence of removing the training
            samples.

        Returns
        -------
        A tensor containing the individual influence scores.
        """
        return self._estimate_influence_value_from_influence_vector(
            self._preprocess_samples(samples_to_evaluate),
            self._compute_influence_vector(train_samples)
        )



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module defining the interface for all the different influence calculator classes.

The BaseInfluenceCalculator interface provides implementations to common to all of
their child classes (i.e. all the different techniques to compute a notion of influence).

It provides optimized implementations for some methods following the assumption that
the computation can be written as a matrix-vector product with a matrix that can be
(pre)-computed and remains unchanged throughout the computation.
"""
from abc import abstractmethod
from enum import Enum
from os import path
from warnings import warn
from xml.dom import NotFoundErr

import tensorflow as tf

from ..utils import BatchSort, BaseNearestNeighbors, LinearNearestNeighbors, ORDER
from ..utils import assert_batched_dataset
from ..types import Optional, Tuple
from ..utils.tf_operations import map_to_device, get_device


class CACHE(Enum):
    """
    Class for the options of where to cache intermediary results for optimizing computations.
    """
    MEMORY = 0
    DISK = 1
    NO_CACHE = 2


class SelfInfluenceCalculator:
    """
    A basic interface for influence calculators whose influence score computation can't be
    decomposed into an inner product between an "influence vector" reflecting the influence
    of the training points and another vector related to a test point. In particular, it will
    be used for RepresenterPointL2 and [WIP] the techniques based on adversarial attacks.
    """

    @abstractmethod
    def _compute_influence_value_from_batch(self, train_samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Computes the influence score (self-influence) for a single batch of training samples.

        Parameters
        ----------
        train_samples
            A tensor with a single batch of training sample.

        Returns
        -------
        influence_values
            The influence score of each sample in the batch train_samples.
        """
        raise NotImplementedError()

    def compute_influence_values(self, train_set: tf.data.Dataset, device: Optional[str] = None) -> tf.data.Dataset:
        """
        Compute the influence score for each sample of the provided (full or partial) model's training dataset.

        If only looking for the values, consider using the utility in deel.influenciae.utils.tf_operations:
        extract_only_values for converting this result into a tensor.

        Parameters
        ----------
        train_set
            A TF dataset with the (full or partial) model's training dataset.
        device
            Device where the computation will be executed

        Returns
        -------
        train_set
            A dataset containing the tuple: (batch of training samples, influence score)
        """
        train_set = map_to_device(train_set, lambda *batch_data: (
            batch_data, self._compute_influence_value_from_batch(batch_data)), device)

        return train_set

    def _compute_influence_values(self, train_set: tf.data.Dataset, device: Optional[str] = None) -> tf.Tensor:
        """
        Compute the influence score for each sample of the provided (full or partial) model's training dataset.
        This version returns a tensor instead of a dataset.
        For internal use only.

        Parameters
        ----------
        train_set
            A TF dataset with the (full or partial) model's training dataset.
        Returns
        -------
        influence score
            A tensor with the sample's influence scores.
        """
        influences_values = self.compute_influence_values(train_set, device)
        influences_values = influences_values.map(
            lambda _, inf_val: inf_val
        )
        inf_val = None
        for batch_inf in influences_values:
            inf_val = batch_inf if inf_val is None else tf.concat([inf_val, batch_inf], axis=0)

        return inf_val

    def compute_top_k_from_training_dataset(
            self,
            train_set: tf.data.Dataset,
            k: int,
            order: ORDER = ORDER.DESCENDING
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        """
        Compute the k most influential data-points of the model's training dataset by computing
        Cook's distance for each point individually.

        Parameters
        ----------
        train_set
            A TF dataset containing the points on which the model was trained.
        k
            An integer with the number of most important samples we wish to keep
        order
            Either ORDER.DESCENDING or ORDER.ASCENDING depending on if we wish to find the top-k or
            bottom-k samples, respectively.

        Returns
        -------
        training_samples, influences_values
            A tuple of tensor.
            - training_samples: A tensor containing the k most influential samples of the training dataset for the model
            provided.
            - influences_values: The influence score corresponding to these k most influential samples.
        """
        assert_batched_dataset(train_set)
        elt_spec = train_set.element_spec[0]
        batch_sorted_dict = BatchSort(elt_spec.shape[1:], (1, k), dtype=elt_spec.dtype, order=order)

        for batch in train_set:
            influence_values = self._compute_influence_value_from_batch(batch)
            if len(influence_values.shape) == 1:
                influence_values = tf.expand_dims(influence_values, axis=-1)
            batch_sorted_dict.add_all(tf.expand_dims(batch[0], axis=0), tf.transpose(influence_values))

        best_samples, best_values = batch_sorted_dict.get()
        influence_values = tf.stack(best_values)
        training_samples = tf.concat(
            [tf.expand_dims(v, axis=0) for v in best_samples], axis=0
        )
        training_samples, influence_values = tf.squeeze(training_samples, axis=0), tf.squeeze(influence_values, axis=0)

        return training_samples, influence_values

    def _save_dataset(self, dataset: tf.data.Dataset, load_or_save_path: str) -> None:
        """
        Save a dataset in the TF dataset format in the specified path.

        Parameters
        ----------
        dataset
            The dataset to save
        load_or_save_path
            The path to save the dataset
        """
        tf.data.experimental.save(dataset, load_or_save_path)

    def _load_dataset(self, dataset_path: str) -> tf.data.Dataset:
        """
        Loads a dataset in the TF format from the specified path.

        Parameters
        ----------
        dataset_path
            The path pointing to the file from which to load the dataset

        Returns
        -------
        dataset
            The target dataset
        """
        if path.exists(dataset_path):
            dataset = tf.data.experimental.load(dataset_path)
        else:
            raise NotFoundErr(f"The dataset path: {dataset_path} was not found")
        return dataset


class BaseInfluenceCalculator(SelfInfluenceCalculator):
    """
    The base implementation of an interface for all the influence calculators in the library.
    All the methods included in deel-influenciae implement these basic functions, with
    each their own notion of influence of a data-point on the model.

    As the computation of the influence scores can be written as an inner product, we apply some
    optimizations that allow us to scale to large datasets.

    Please note that for some of the classes implementing this interface, the notion of influence
    vector does not necessarily have the meaning of being the delta of the weights of the model
    after the perturbation of the training dataset.
    """

    @abstractmethod
    def _preprocess_samples(self, samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Preprocess a sample to evaluate

        Parameters
        ----------
        samples
            sample to evaluate
        Returns
        -------
        The preprocessed sample to evaluate
        """
        raise NotImplementedError()

    @abstractmethod
    def _compute_influence_vector(self, train_samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Computes the influence vector (i.e. the delta of model's weights after a perturbation on the training
        dataset) for a single batch of training samples.

        Parameters
        ----------
        train_samples
            A tuple with the batch of training samples (with their labels).

        Returns
        -------
        influence_vector
            A tensor with the influence vector for each individual point.
        """
        raise NotImplementedError()

    def compute_influence_vector(
            self,
            train_set: tf.data.Dataset,
            save_influence_vector_ds_path: Optional[str] = None,
            device: Optional[str] = None
    ) -> tf.data.Dataset:
        """
        Compute the influence vector for each sample of the provided (full or partial) model's training dataset.

        Parameters
        ----------
        train_set
            A TF dataset with the (full or partial) model's training dataset.
        save_influence_vector_ds_path
            The path to save or load the influence vector of the training dataset. If specified,
            load the dataset if it has already been computed, otherwise, compute the influence vector and
            then save it in the specified path.
        device
            Device where the computation will be executed

        Returns
        -------
        inf_vect_ds
            A dataset containing the tuple: (batch of training samples, influence vector)
        """
        inf_vect_ds = map_to_device(train_set, lambda *batch: (batch, self._compute_influence_vector(batch)), device)
        if save_influence_vector_ds_path is not None:
            inf_vect = inf_vect_ds.map(lambda *batch: batch[-1])
            self._save_dataset(inf_vect.unbatch(), save_influence_vector_ds_path)

        return inf_vect_ds

    def estimate_influence_values_in_batches(
            self,
            dataset_to_evaluate: tf.data.Dataset,
            train_set: tf.data.Dataset,
            influence_vector_in_cache: CACHE = CACHE.MEMORY,
            load_influence_vector_path: Optional[str] = None,
            save_influence_vector_path: Optional[str] = None,
            save_influence_value_path: Optional[str] = None,
            device: Optional[str] = None
    ) -> tf.data.Dataset:
        """
        Estimates the influence that each point in the provided training dataset has on each of the test points.
        This can provide some insights as to what makes the model predict a certain way for the given test points,
        and thus presents data-centric explanations.

        Parameters
        ----------
        dataset_to_evaluate
            A TF dataset containing the test samples for which to compute the effect of removing each of the provided
            training points (individually).
        train_set
            A TF dataset containing the model's training dataset (partial or full).
        influence_vector_in_cache
            An enum indicating if intermediary values are to be cached (either in memory or on the disk) or not.
            Options include CACHE.MEMORY (0) for caching in memory, CACHE.DISK (1) for the disk and CACHE.NO_CACHE (2)
            for no optimization.
        load_influence_vector_path
            The path to load the influence vectors (if they have already been calculated).
        save_influence_vector_path
            The path to save the computed influence vector.
        save_influence_value_path
            The path to save the computed influence values.
        device
            Device where the computation will be executed

        Returns
        -------
        influence_value_dataset
            A dataset containing the tuple: (samples_to_evaluate, dataset).

            - samples_to_evaluate: The batch of sample to evaluate.
            - dataset: Dataset containing tuples of batch of the training dataset and their influence score.
        """
        if not influence_vector_in_cache and load_influence_vector_path is None:
            warn("Warning: The computation is not efficient, thinks to use cache or disk save")

        if influence_vector_in_cache == CACHE.MEMORY:
            load_influence_vector_path = None

        if load_influence_vector_path is not None and influence_vector_in_cache == CACHE.DISK:
            inf_vect_ds = self._load_dataset(load_influence_vector_path)
            batch_size = train_set._batch_size  # pylint: disable=W0212
            inf_vect_ds = tf.data.Dataset.zip((train_set, inf_vect_ds.batch(batch_size)))
        else:
            inf_vect_ds = self.compute_influence_vector(train_set, save_influence_vector_path, device)

        if influence_vector_in_cache == CACHE.MEMORY:
            inf_vect_ds = inf_vect_ds.cache()

        influence_value_dataset = map_to_device(dataset_to_evaluate,
                                                lambda *batch_evaluate: self._estimate_inf_values_with_inf_vect_dataset(
                                                    inf_vect_ds, batch_evaluate), device)

        if save_influence_value_path is not None:
            for batch_idx, (_, samples_inf_val_dataset) in enumerate(influence_value_dataset):
                self._save_dataset(samples_inf_val_dataset, f"{save_influence_value_path}/batch_{batch_idx:06d}")

        return influence_value_dataset

    def top_k(  # pylint: disable=R0913
            self,
            dataset_to_evaluate: tf.data.Dataset,
            train_set: tf.data.Dataset,
            k: int = 5,
            nearest_neighbors: BaseNearestNeighbors = LinearNearestNeighbors(),
            influence_vector_in_cache: CACHE = CACHE.MEMORY,
            load_influence_vector_ds_path: Optional[str] = None,
            save_influence_vector_ds_path: Optional[str] = None,
            save_top_k_ds_path: Optional[str] = None,
            order: ORDER = ORDER.DESCENDING,
            d_type: tf.DType = tf.float32,
            device: Optional[str] = None
    ) -> tf.data.Dataset:
        """
        Find the top-k closest elements for each element of dataset to evaluate in the training dataset
        The method will return a dataset containing a tuple of:
            (Top-k influence values for each sample to evaluate, Top-k training sample for each sample to evaluate)

        Parameters
        ----------
        dataset_to_evaluate
            The dataset which contains the samples which will be compare to the training dataset
        train_set
            The dataset used to train the model.
        k
            the number of most influence samples to retain in training dataset
        nearest_neighbors
            The nearest neighbor method. The default method is a linear search
        influence_vector_in_cache
            An enum indicating if intermediary values are to be cached (either in memory or on the disk) or not.
            Options include CACHE.MEMORY (0) for caching in memory, CACHE.DISK (1) for the disk and CACHE.NO_CACHE (2)
            for no optimization.
        load_influence_vector_ds_path
            The path to load the influence vectors (if they have already been calculated).
        save_influence_vector_ds_path
            The path to save the computed influence vector.
        save_top_k_ds_path
            The path to save the result of the computation of the top-k elements
        order
            Either ORDER.DESCENDING or ORDER.ASCENDING depending on if we wish to find the top-k or
            bottom-k samples, respectively.
        d_type
            The data-type of the tensors.
        device
            Device where the computation will be executed

        Returns
        -------
        top_k_dataset
            A dataset containing the tuple (samples_to_evaluate, influence_values, training_samples).

            - samples_to_evaluate: Top-k samples to evaluate.
            - influence_values: Top-k influence values for each sample to evaluate.
            - training_samples: Top-k training sample for each sample to evaluate.
        """
        if not influence_vector_in_cache and load_influence_vector_ds_path is None:
            warn("Warning: The computation is not efficient thinks to use cache or disk save")

        if influence_vector_in_cache == CACHE.MEMORY:
            load_influence_vector_ds_path = None

        if load_influence_vector_ds_path is not None and influence_vector_in_cache == CACHE.DISK:
            inf_vect_ds = self._load_dataset(load_influence_vector_ds_path)
            batch_size = train_set._batch_size  # pylint: disable=W0212
            inf_vect_ds = tf.data.Dataset.zip((train_set, inf_vect_ds.batch(batch_size)))
        else:
            inf_vect_ds = self.compute_influence_vector(train_set, save_influence_vector_ds_path, device)

        if influence_vector_in_cache == CACHE.MEMORY:
            inf_vect_ds = inf_vect_ds.cache()

        batch_size_eval = int(dataset_to_evaluate._batch_size)  # pylint: disable=W0212
        nearest_neighbors.build(
            inf_vect_ds,
            self._estimate_influence_value_from_influence_vector,
            k,
            query_batch_size=batch_size_eval,
            d_type=d_type,
            order=order,
        )

        top_k_dataset = dataset_to_evaluate.map(
            lambda *batch_evaluate: self._top_k_with_inf_vect_dataset_train(
                batch_evaluate, nearest_neighbors, batch_size_eval, device
            )
        )

        if save_top_k_ds_path is not None:
            self._save_dataset(top_k_dataset, save_top_k_ds_path)

        return top_k_dataset

    def _estimate_inf_values_with_inf_vect_dataset(
            self,
            inf_vect_dataset: tf.data.Dataset,
            samples_to_evaluate: Tuple[tf.Tensor, ...],
            device: Optional[str] = None
    ) -> Tuple[Tuple[tf.Tensor, ...], tf.data.Dataset]:
        """
        Internal function to optimize computations when the influence vectors have already been calculated.

        Estimates the influence score between the samples we wish to evaluate and the set of influence
        vectors from the training dataset.

        Parameters
        ----------
        inf_vect_dataset
            A TF dataset with the influence vectors computed using some of the model's training data-points.
        samples_to_evaluate
            A tensor containing a single batch of samples of which we wish to estimate the influence of
             leaving out the training points corresponding to the influence vectors.
        device
            Device where the computation will be executed
        Returns
        -------
        A dataset containing the tuple:
            batch of the training dataset
            influence scores
        """
        samples_inf_val_dataset = map_to_device(inf_vect_dataset, lambda *batch:
                    (batch[:-1][0],
                     self._estimate_influence_values_from_influence_vector(samples_to_evaluate, batch[-1])), device)
        return samples_to_evaluate, samples_inf_val_dataset

    def _top_k_with_inf_vect_dataset_train(
            self,
            sample_to_evaluate: Tuple[tf.Tensor, ...],
            nearest_neighbor: BaseNearestNeighbors,
            batch_size_eval: Optional[int] = None,
            device: Optional[str] = None
    ) -> Tuple[Tuple[tf.Tensor, ...], tf.Tensor, Tuple[tf.Tensor, ...]]:
        """
        Internal function to optimize computations when the influence vectors have already been calculated.

        Finds the top-k most influential training points for each test sample.

        Parameters
        ----------
        sample_to_evaluate
            A batched tensor containing the samples which will be compare to the training dataset
        nearest_neighbor
            The nearest neighbor method
        k
            the number of most influence samples to retain in training dataset
        device
            Device where the computation will be executed
        Returns
        -------
        sample_to_evaluate
            sample to evaluate
        influence_values
            Top-k influence values for each sample to evaluate.
        training_samples
            Top-k training sample for each sample to evaluate.
        """
        device = get_device(device)
        with tf.device(device):
            v_to_evaluate = self._preprocess_samples(sample_to_evaluate)
            if batch_size_eval is None:
                influences_values, training_samples = nearest_neighbor.query(v_to_evaluate)
            else:
                influences_values, training_samples = nearest_neighbor.query(v_to_evaluate, batch_size_eval)

        return sample_to_evaluate, influences_values, training_samples

    def _estimate_influence_values_from_influence_vector(
            self,
            samples_to_evaluate: Tuple[tf.Tensor, ...],
            inf_vect: tf.Tensor
    ) -> tf.Tensor:
        """
        Internal function to optimize computations when the influence vectors have already been calculated.

        Computes the influence values between each of the test samples and one influence vector from the training
        dataset.

        Parameters
        ----------
        samples_to_evaluate
            A single batch of test samples for which we wish to compute the influence of leaving out the training
            data-points corresponding to the influence vector.
        inf_vect
            A tensor with one influence vector

        Returns
        -------
        Tuple:
            batch of the training dataset
            influence vector
        """
        v_to_evaluate = self._preprocess_samples(samples_to_evaluate)
        value = self._estimate_influence_value_from_influence_vector(v_to_evaluate, inf_vect)

        return value

    @tf.function
    def _estimate_individual_influence_values_from_batch(
            self,
            train_samples: Tuple[tf.Tensor, ...],
            samples_to_evaluate: Tuple[tf.Tensor, ...]
    ) -> tf.Tensor:
        """
        Estimates the influence value of leaving out a single training sample on the provided test sample.

        Parameters
        ----------
        train_samples
            A single training sample
        samples_to_evaluate
            A single test sample.

        Returns
        -------
        influence_values
            A tensor with the resulting influence value.
        """
        v_train = self._compute_influence_vector(train_samples)
        v_to_evaluate = self._preprocess_samples(samples_to_evaluate)
        influence_values = self._estimate_influence_value_from_influence_vector(v_to_evaluate, v_train)

        return influence_values

    @abstractmethod
    def _estimate_influence_value_from_influence_vector(
            self,
            preproc_test_sample: tf.Tensor,
            influence_vector: tf.Tensor
    ) -> tf.Tensor:
        """
        Estimates the influence score of leaving out the influence vector corresponding to a given training
        data-point on a test sample that has already been pre-processed.

        Parameters
        ----------
        preproc_test_sample
            A single pre-processed test sample we wish to evaluate.
        influence_vector
            A single influence vector corresponding to a data-point from the training dataset.

        Returns
        -------
        influence_values
            A tensor with the resulting influence value.
        """
        raise NotImplementedError()



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Influence function module
"""
from .first_order_influence_calculator import FirstOrderInfluenceCalculator
from .second_order_influence_calculator import SecondOrderInfluenceCalculator
from .arnoldi_influence_calculator import ArnoldiInfluenceCalculator
from .base_group_influence import BaseGroupInfluenceCalculator



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module implementing the representer point theorem for kernels for estimating the
influence of training data-points, as per:
https://arxiv.org/abs/1811.09720
"""
import tensorflow as tf
from tensorflow.keras import Model #pylint:  disable=E0611
from tensorflow.keras.layers import Input, Dense #pylint:  disable=E0611
from tensorflow.keras.losses import MeanSquaredError, Loss #pylint:  disable=E0611
from tensorflow.keras.regularizers import L2 #pylint:  disable=E0611

from .base_representer_point import BaseRepresenterPoint
from ..types import Tuple, Callable, Union

from ..utils import BacktrackingLineSearch, dataset_size


class RepresenterPointL2(BaseRepresenterPoint):
    """
    A class implementing a method to compute the influence of training points through
    the representer point theorem for kernels.

    It builds a kernel that approximates the model's last layer such that for a training
    point x_i and a test point x_t with label y_t:

    y_t = sum_i k(alpha_i, x_i, x_t)

    Disclaimer: This method only works on classification problems!

    Parameters
    ----------
    model
        A TF2 model that has already been trained
    train_set
        A batched TF dataset with the points with which the model was trained
    loss_function
        The loss function with which the model was trained. This loss function MUST NOT be reduced.
    lambda_regularization
        The coefficient for the regularization of the surrogate last layer that needs
        to be trained for this method
    scaling_factor
        A float with the scaling factor for the SGD backtracking line-search optimizer
        for fitting the surrogate linear model
    epochs
        An integer for the amount of epochs to fit the linear model
    layer_index
        layer of the logits
    """

    def __init__(
            self,
            model: Model,
            train_set: tf.data.Dataset,
            loss_function: Union[Callable[[tf.Tensor, tf.Tensor], tf.Tensor], Loss],
            lambda_regularization: float,
            scaling_factor: float = 0.1,
            epochs: int = 100,
            layer_index: int = -1,
    ):
        super().__init__(model, train_set, loss_function, layer_index)
        self.n_train = dataset_size(train_set)
        self.train_set = train_set
        self.lambda_regularization = lambda_regularization
        self.scaling_factor = scaling_factor
        self.epochs = epochs
        self.linear_layer = None
        self._train_last_layer(self.epochs)

    def _train_last_layer(self, epochs: int):
        """
        Trains an L2-regularized surrogate linear model to predict like the model on the
        training dataset. The optimization is done using a Backtracking Line-Search
        algorithm with the Armijo condition and SGD as the optimizer as it was done
        in the original implementation.

        Parameters
        ----------
        epochs
            An integer with the amount of epochs to train the surrogate model
        """
        self.linear_layer = self._create_surrogate_model()
        optimizer = BacktrackingLineSearch(batches_per_epoch=self.n_train / self.train_set._batch_size,  # pylint: disable=W0212
                                           scaling_factor=self.scaling_factor)  # the optimizer used in the paper's code
        loss_function = self.loss_function
        mse_loss = MeanSquaredError(reduction=tf.keras.losses.Reduction.SUM_OVER_BATCH_SIZE)

        self.linear_layer.compile(optimizer=optimizer, loss=mse_loss)
        for _ in range(epochs):
            for x_batch, _ in self.train_set:
                loss, grads, z_batch, y_target = self._learn_step_last_layer(x_batch, mse_loss)
                optimizer.step(self.linear_layer, loss, z_batch, y_target, grads)

        self.linear_layer.compile(optimizer=optimizer, loss=loss_function)

    @tf.function
    def _learn_step_last_layer(
            self,
            x_batch: tf.Tensor,
            mse_loss: Callable[[tf.Tensor, tf.Tensor], tf.Tensor]
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
        """
        Trains the L2-regularized surrogate linear model to predict like the model on the training
        dataset for one optimizer step on a single batch.

        Parameters
        ----------
        x_batch
            A training sample wrt to which we wish to compute the gradients
        mse_loss
            A callable that computes the MSE loss

        Returns
        -------
        A tuple with (value of the loss, gradients of the linear model, the latent space of the batch, the prediction)
        """
        z_batch = self.feature_extractor(x_batch)
        y_target = self.model.layers[-1](z_batch)
        with tf.GradientTape() as tape:
            logits = self.linear_layer(z_batch, training=True)
            loss = mse_loss(y_target, logits)
        gradients = tape.gradient(loss, self.linear_layer.trainable_weights)
        return loss, gradients, z_batch, y_target

    def _create_surrogate_model(self) -> Model:
        """
        Instances an L2-regularized linear model to use as surrogate with the
        right input and output shapes.

        Returns
        -------
        surrogate_model
            A TF2 L2-regularized linear model
        """
        inputs = Input(shape=self.feature_extractor.output_shape[1:], dtype=self.model.output.dtype)
        last_layer = Dense(self.model.output_shape[-1], use_bias=False,
                           kernel_regularizer=L2(self.lambda_regularization),
                           dtype=self.model.output.dtype)
        outputs = last_layer(inputs)
        surrogate_model = Model(inputs=inputs, outputs=outputs)
        surrogate_model.layers[-1].trainable = True
        surrogate_model.compile(loss=self.model.compiled_loss)

        return surrogate_model

    def _compute_alpha(self, z_batch: tf.Tensor, y_batch: tf.Tensor) -> tf.Tensor:
        """
        Computes the alpha factor for the kernel approximation. This element gives a notion of
        the resistance that each training data-point towards minimizing the norm of the linear
        layer's weight matrix. This is essentially this method's notion of influence score.

        Parameters
        ----------
        z_batch
            a training sample wrt to which we wish to compute the gradients
        y_batch
            label associated to the training sample

        Returns
        -------
        alpha
            mean of the gradient
        """
        with tf.GradientTape(persistent=False, watch_accessed_variables=False) as tape:
            tape.watch(self.linear_layer.weights)
            logits = self.linear_layer(z_batch)
            loss = self.linear_layer.compiled_loss(y_batch, logits)
        alpha = tape.jacobian(loss, self.linear_layer.weights)[0]
        alpha = tf.divide(
            alpha,
            -2. * self.lambda_regularization * tf.cast(self.n_train, alpha.dtype) + tf.constant(1e-5, dtype=alpha.dtype)
        )

        # Now, divide each of the alpha_i by their feature maps
        alpha = tf.multiply(
            alpha,
            tf.repeat(
                tf.expand_dims(
                    tf.divide(tf.ones_like(z_batch), z_batch + tf.constant(1e-5, dtype=alpha.dtype)),
                    axis=-1),
                alpha.shape[-1], axis=-1
            )
        )  # Do the multiplication part of the inner product
        alpha = tf.reduce_sum(alpha, axis=1)  # Now do the sum

        return alpha

    def predict_with_kernel(self, samples_to_evaluate: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Uses the learned kernel to approximate the model's predictions on a group of samples.

        Parameters
        ----------
        samples_to_evaluate
            A single batch of tensors with the samples for which we wish to approximate the model's
            predictions

        Returns
        -------
        predictions
            A tensor with an approximation of the model's predictions
        """
        influence_vectors = self.compute_influence_vector(self.train_set)
        _, dataset_influence = self._estimate_inf_values_with_inf_vect_dataset(influence_vectors, samples_to_evaluate)
        dataset_influence = dataset_influence.map(lambda x, v: v)
        dataset_iterator = iter(dataset_influence)

        def body_fun(i, value):
            v = next(dataset_iterator)
            i = i + 1
            value = tf.cast(value, v.dtype) + tf.reduce_sum(v, axis=1)
            return i, value

        _, predictions = tf.while_loop(lambda i, value: i < dataset_influence.cardinality(), body_fun,
                                            [tf.constant(0, dtype=tf.int64),
                                             tf.zeros((tf.shape(samples_to_evaluate[-1])[0],), dtype=tf.float32)])

        return predictions



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Conjugate Gradients solver based on jax.scipy's implementation and wikipedia
https://jax.readthedocs.io/en/latest/_autosummary/jax.scipy.sparse.linalg.cg.html
https://en.wikipedia.org/wiki/Conjugate_gradient_method

BiCGSTAB (Biconjugate Gradient Stabilized) solver based also on jax.scipy's implementation and wikipedia
https://jax.readthedocs.io/en/latest/_autosummary/jax.scipy.sparse.linalg.bicgstab.html
https://en.wikipedia.org/wiki/Biconjugate_gradient_stabilized_method#Preconditioned_BiCGSTAB
"""
import tensorflow as tf

from ..types import Callable, Optional


def _identity(x): # pylint: disable=C0116
    return x


def conjugate_gradients_solve(
        operator: Callable,
        b: tf.Tensor,
        x0: Optional[tf.Tensor] = None,
        *,
        maxiter: int,
        tol: float = 1e-3,
        atol: float = 1e-5,
        M: Callable = _identity
):
    """
    A simple Conjugate Gradients solver based on jax.scipy

    Parameters
    ----------
    operator: Callable
        The operator that calculates the linear map A(x). It is assumed to be hermitian and positive definite
    b: tf.Tensor
        The right hand side of the linear system, represented by a single vector.
    x0: Optional
        A tensor with the same shape as b and the output that servers as a first guess for the solution
    maxiter: int
        The maximum amount of iterations
    tol: float
        Tolerance for convergence. norm(residual) <= max(tol * norm(b), atol)
    atol: float
        Tolerance for convergence. norm(residual) <= max(tol * norm(b), atol)
    M: Callable
        A preconditioner approximating the inverse of A.

    Returns
    -------
    x_final: tf.Tensor
        A tensor with the solution found by the solver
    """
    if x0 is None:
        x0 = tf.zeros_like(b)

    bs = tf.reduce_sum(tf.matmul(b, b, transpose_a=True))
    atol2 = tf.reduce_max([tf.cast(tf.square(tol), dtype=bs.dtype) * bs, tf.cast(tf.square(atol), dtype=bs.dtype)])

    def cond_fun(_, r, gamma, __, k):
        rs = gamma if M is _identity else tf.reduce_sum(tf.matmul(r, r, transpose_a=True))
        cond1 = tf.greater(rs, atol2)
        cond2 = tf.greater(maxiter, k)
        return tf.logical_and(cond1, cond2)

    def body_fun(x, r, gamma, p, k):
        Ap = operator(p)
        alpha = gamma / (tf.reduce_sum(tf.matmul(p, Ap, transpose_a=True)))
        x_ = x + alpha * p
        r_ = r - alpha * Ap
        z_ = M(r_)
        gamma_ = tf.reduce_sum(tf.matmul(r_, z_, transpose_a=True))
        beta_ = gamma_ / gamma
        p_ = z_ + beta_ * p

        return x_, r_, gamma_, p_, k + 1

    r0 = b - operator(x0)
    p0 = z0 = M(r0)
    gamma0 = tf.reduce_sum(tf.matmul(r0, z0, transpose_a=True))
    initial_value = [x0, r0, gamma0, p0, tf.constant(0, dtype=tf.int32)]

    val = tf.while_loop(
        cond=cond_fun,
        body=body_fun,
        loop_vars=initial_value
    )

    x_final, *_ = val

    return x_final


def biconjugate_gradient_stabilized_solve(
        operator: Callable,
        b: tf.Tensor,
        x0: Optional[tf.Tensor] = None,
        *,
        maxiter: int,
        tol: float = 1e-5,
        atol: float = 1e-6,
        M: Callable = _identity):
    """
    A simple BiCGSTAB solver based on jax.scipy

    Parameters
    ----------
    operator: Callable
        The operator that calculates the linear map A(x). It is assumed to be hermitian and positive definite
    b: tf.Tensor
        The right hand side of the linear system, represented by a single vector.
    x0: Optional
        A tensor with the same shape as b and the output that servers as a first guess for the solution
    maxiter: int
        The maximum amount of iterations
    tol: float
        Tolerance for convergence. norm(residual) <= max(tol * norm(b), atol)
    atol: float
        Tolerance for convergence. norm(residual) <= max(tol * norm(b), atol)
    M: Callable
        A preconditioner approximating the inverse of A.

    Returns
    -------
    x_final: tf.Tensor
        A tensor with the solution found by the solver
    """
    if x0 is None:
        x0 = tf.zeros_like(b)

    bs = tf.reduce_sum(tf.matmul(b, b, transpose_a=True))
    atol2 = tf.reduce_max([tf.square(tol) * bs, tf.square(atol)])

    def cond_fun(value):
        _, r, *_, k = value
        rs = tf.reduce_sum(tf.matmul(r, r, transpose_a=True))
        return (rs > atol2) & (k < maxiter) & (k >= 0)

    def body_fun(value):
        x, r, rhat, alpha, omega, rho, p, q, k = value
        rho_ = tf.reduce_sum(tf.matmul(rhat, r, transpose_a=True))
        beta = rho_ / rho * alpha / omega
        p_ = r + beta * (p - omega * q)
        phat = M(p_)
        q_ = operator(phat)
        alpha_ = rho_ / tf.reduce_sum(tf.matmul(rhat, q_, transpose_a=True))
        s = r - alpha_ * q_
        exit_early = tf.reduce_sum(tf.matmul(s, s, transpose_a=True)) < atol2
        shat = M(s)
        t = operator(shat)
        omega_ = tf.reduce_sum(tf.matmul(t, s, transpose_a=True)) / tf.reduce_sum(tf.matmul(t, t, transpose_a=True))
        x_ = x + alpha_ * phat if exit_early else x + alpha_ * phat + omega_ * shat
        r_ = s if exit_early else s - omega_ * t
        k_ = tf.where((omega_ == 0) | (alpha_ == 0), -11, k + 1)
        k_ = tf.where((rho_ == 0), -10, k_)
        return x_, r_, rhat, alpha_, omega_, rho_, p_, q_, k_

    r0 = b - operator(x0)
    rho0 = alpha0 = omega0 = 1.
    initial_value = (x0, r0, r0, alpha0, omega0, rho0, r0, r0, 0)

    # Perform the minimization until convergence
    val = initial_value
    while cond_fun(val):
        val = body_fun(val)
    x_final, *_ = val

    return x_final
